# Copyright (c) 2018 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml
import io
import requests
import zipfile
import unicodedata
import os
from time import strptime

CITATION_FILE_VERSION_FIELD = "version"
CITATION_FILE_DATE_FIELD = "date-released"
CITATION_AUTHORS_TYPE = "authors"
CITATION_AUTHOR_FIRST_NAME = "given-names"
CITATION_AUTHOR_SURNAME = "family-names"
CITATION_FILE_DESCRIPTION = "title"

ZENODO_RELATION_FIELD = "relation"
ZENODO_NEWER_VERSION_OF = 'isNewVersionOf'
ZENODO_SIBLING_OF = "cites"
ZENODO_METADATA = 'metadata'
ZENODO_PRE_RESERVED_DOI = "prereserve_doi"
ZENODO_DOI_VALUE = "doi"
ZENODO_DEPOSIT_ID = 'id'
ZENODO_METADATA_TITLE = "title"
ZENODO_METATDATA_DESC = "description"
ZENODO_METADATA_CREATORS = "creators"
ZENODO_FILE = "file"
ZENODO_AUTHOR_NAME = "name"

AUTHOR_AFFILIATION = "affiliation"
AUTHOR_ORCID = "orcid"
IDENTIFIER = 'identifier'

# pylint: skip-file


class _ZenodoException(Exception):
    """
    Exception from a call to Zenodo.
    """

    def __init__(self, operation, expected, request):
        super().__init__(
            "don't know what went wrong. got wrong status code when trying "
            f"to {operation}. Got error code {request.status_code} (when "
            f"expecting {expected}) with response content {request.content}")
        self.request = request
        self.expected = expected


class _Zenodo(object):
    """
    Manages low level access to Zenodo.
    """

    # pragma: no cover
    __slots__ = ("__zenodo_token", )

    _BASE_URI = "https://zenodo.org/api"
    _DEPOSIT_GET_URL = _BASE_URI + "/deposit/depositions"
    _DEPOSIT_PUT_URL = _BASE_URI + "/deposit/depositions/{}/files"
    _PUBLISH_URL = _BASE_URI + "/deposit/depositions/{}/actions/publish"
    _CONTENT_TYPE = "Content-Type"
    _JSON = "application/json"
    _ACCESS_TOKEN = 'access_token'
    _RELATED_IDENTIFIERS = 'related_identifiers'
    _VALID_STATUS_REQUEST_GET = 200
    _VALID_STATUS_REQUEST_POST = 201
    _VALID_STATUS_REQUEST_PUBLISH = 202

    def __init__(self, token):
        self.__zenodo_token = token

    @staticmethod
    def _json(r):
        try:
            return r.json()
        except Exception:  # pylint: disable=broad-except
            return None

    def get_verify(self, related):
        r = requests.get(
            self._DEPOSIT_GET_URL, timeout=10,
            params={self._ACCESS_TOKEN: self.__zenodo_token,
                    self._RELATED_IDENTIFIERS: related},
            json={}, headers={self._CONTENT_TYPE: self._JSON})
        if r.status_code != self._VALID_STATUS_REQUEST_GET:
            raise _ZenodoException(
                "request a DOI", self._VALID_STATUS_REQUEST_GET, r)
        return self._json(r)

    def post_create(self, related):
        r = requests.post(
            self._DEPOSIT_GET_URL, timeout=10,
            params={self._ACCESS_TOKEN: self.__zenodo_token,
                    self._RELATED_IDENTIFIERS: related},
            json={}, headers={self._CONTENT_TYPE: self._JSON})
        if r.status_code != self._VALID_STATUS_REQUEST_POST:
            raise _ZenodoException(
                "get an empty upload", self._VALID_STATUS_REQUEST_POST, r)
        return self._json(r)

    def post_upload(self, deposit_id, data, files):
        r = requests.post(
            self._DEPOSIT_PUT_URL.format(deposit_id), timeout=10,
            params={self._ACCESS_TOKEN: self.__zenodo_token},
            data=data, files=files)
        if r.status_code != self._VALID_STATUS_REQUEST_POST:
            raise _ZenodoException(
                "to put files and data into the preallocated DOI",
                self._VALID_STATUS_REQUEST_POST, r)
        return self._json(r)

    def post_publish(self, deposit_id):
        r = requests.post(
            self._PUBLISH_URL.format(deposit_id), timeout=10,
            params={self._ACCESS_TOKEN: self.__zenodo_token})
        if r.status_code != self._VALID_STATUS_REQUEST_PUBLISH:
            raise _ZenodoException(
                "publish the DOI", self._VALID_STATUS_REQUEST_PUBLISH, r)
        return self._json(r)


class CitationUpdaterAndDoiGenerator(object):
    def __init__(self):
        self.__zenodo = None

    def update_citation_file_and_create_doi(
            self, citation_file_path, doi_title, create_doi, publish_doi,
            previous_doi, zenodo_access_token, module_path):
        """
        Take a CITATION.cff file and updates the version and
        date-released fields, and rewrites the ``CITATION.cff`` file.

        :param str citation_file_path: File path to the ``CITATION.cff`` file
        :param bool create_doi:
            Whether to use Zenodo DOI interface to grab a DOI
        :param str zenodo_access_token: Access token for Zenodo
        :param bool publish_doi: Whether to publish the DOI on Zenodo
        :param str previous_doi: DOI to append the created DOI to
        :param str doi_title: Title for the created DOI
        :param str module_path: Path to the module to zip up
        :param bool update_version:
            Whether we should update the citation version
        """
        self.__zenodo = _Zenodo(zenodo_access_token)

        # data holders
        deposit_id = None

        # read in YAML file
        with open(citation_file_path, 'r', encoding="utf-8") as stream:
            yaml_file = yaml.safe_load(stream)

        # if creating a DOI, go and request one
        if create_doi:
            doi_id, deposit_id = self._request_doi(previous_doi)
            yaml_file[IDENTIFIER] = doi_id

        # rewrite citation file with updated fields
        with io.open(citation_file_path, 'w', encoding='utf8') as outfile:
            yaml.dump(yaml_file, outfile, default_flow_style=False,
                      allow_unicode=True)

        # if creating a DOI, finish the request and possibly publish it
        if create_doi:
            self._finish_doi(
                deposit_id, publish_doi, doi_title,
                yaml_file[CITATION_FILE_DESCRIPTION], yaml_file, module_path)

    def _request_doi(self, previous_doi):
        """
        Go to Zenodo and requests a DOI.

        :param str previous_doi: the previous DOI for this module, if exists
        :return: the DOI id, and deposit id
        :rtype: tuple(str, str)
        """
        # create link to previous version (if applicable)
        related = list()
        related.append({
            ZENODO_RELATION_FIELD: ZENODO_NEWER_VERSION_OF,
            IDENTIFIER: previous_doi})

        # get a request for a DOI
        self.__zenodo.get_verify(related)

        # get empty upload
        request_data = self.__zenodo.post_create(related)

        # get DOI and deposit id
        doi_id = unicodedata.normalize(
            'NFKD',
            (request_data[ZENODO_METADATA][ZENODO_PRE_RESERVED_DOI]
             [ZENODO_DOI_VALUE])).encode('ascii', 'ignore')
        deposition_id = request_data[ZENODO_DEPOSIT_ID]

        return doi_id, deposition_id

    def _finish_doi(
            self, deposit_id, publish_doi, title,
            doi_description, yaml_file, module_path):
        """
        Finishes the DOI on Zenodo.

        :param str deposit_id: the deposit id to publish
        :param bool publish_doi: whether we should publish the DOI
        :param str title: the title of this DOI
        :param str doi_description: the description for the DOI
        :param yaml_file: the citation file after its been read it
        :param module_path: the path to the module to DOI
        """
        zipped_file = None
        try:
            zipped_file = self._zip_up_module(module_path)
            with open(zipped_file, "rb") as zipped_open_file:
                files = {ZENODO_FILE: zipped_open_file}
                data = self._fill_in_data(title, doi_description, yaml_file)
                self.__zenodo.post_upload(deposit_id, data, files)
        finally:
            if zipped_file:
                os.remove(zipped_file)

        # publish DOI
        if publish_doi:
            self.__zenodo.post_publish(deposit_id)

    def _zip_up_module(self, module_path):
        """
        Zip up a module.

        :param str module_path: the path to the module to zip up
        :return: the filename to the zip file
        """
        if os.path.isfile('module.zip'):
            os.remove('module.zip')

        avoids = [".git", ".gitignore", ".gitattributes", ".travis.yml",
                  ".github", "model_binaries", "common_model_binaries",
                  ".coveragerc", ".idea"]

        with zipfile.ZipFile(
                'module.zip', 'w', zipfile.ZIP_DEFLATED) as module_zip_file:
            self._zip_walker(module_path, avoids, module_zip_file)
        return 'module.zip'

    @staticmethod
    def _zip_walker(module_path, avoids, module_zip_file):
        """
        Traverse the module and its sub-directories and only adds to the
        files to the zip which are not within a avoid directory that.

        :param str module_path: the path to start the search at
        :param set(str) avoids: the set of avoids to avoid
        :param ~zipfile.ZipFile module_zip_file: the zip file to put into
        """
        for directory_path, _, files in os.walk(module_path):
            for directory_name in directory_path.split(os.sep):
                if directory_name in avoids:
                    break
            else:
                for potential_zip_file in files:
                    # if safe to zip, zip
                    if potential_zip_file not in avoids:
                        module_zip_file.write(
                            os.path.join(directory_path, potential_zip_file))

    @staticmethod
    def _fill_in_data(doi_title, doi_description, yaml_file):
        """
        Add in data to the Zenodo metadata.

        :param str doi_title: the title of the DOI
        :param str doi_description: the description of the DOI
        :param dict yaml_file: the citation file once read into the system
        :return: dict containing Zenodo metadata
        :rtype: dict
        """
        # add basic meta data
        metadata = {
            ZENODO_METADATA_TITLE: doi_title,
            ZENODO_METATDATA_DESC: doi_description,
            ZENODO_METADATA_CREATORS: []
        }

        # get author data from the citation file
        for author in yaml_file[CITATION_AUTHORS_TYPE]:
            author_data = {
                ZENODO_AUTHOR_NAME: (
                    author[CITATION_AUTHOR_SURNAME] + ", " +
                    author[CITATION_AUTHOR_FIRST_NAME])
            }
            if AUTHOR_AFFILIATION in author:
                author_data[AUTHOR_AFFILIATION] = author[AUTHOR_AFFILIATION]
            if AUTHOR_ORCID in author:
                author_data[AUTHOR_ORCID] = author[AUTHOR_ORCID]
            metadata[ZENODO_METADATA_CREATORS].append(author_data)
        return {ZENODO_METADATA: metadata}

    @staticmethod
    def convert_text_date_to_date(
            version_month, version_year, version_day):
        """
        Convert the 3 components of a date into a CFF date.

        :param version_month: version month, in text form
        :type version_month: str or int
        :param int version_year: version year
        :param int version_day: version day of month
        :return: the string representation for the CFF file
        :rtype: str
        """
        return "{}-{}-{}".format(
            version_year,
            CitationUpdaterAndDoiGenerator.convert_month_name_to_number(
                version_month),
            version_day)

    @staticmethod
    def convert_month_name_to_number(version_month):
        """
        Convert a python month in text form to a number form.

        :param version_month: the text form of the month
        :type version_month: str or int
        :return: the month int value
        :rtype: int
        :raises ValueError: when the month name is not recognised
        """
        if isinstance(version_month, int):
            return version_month
        elif isinstance(version_month, str):
            try:
                return int(version_month)
            except ValueError as original:
                try:
                    return strptime(version_month, "%B").tm_mon
                except ValueError:
                    try:
                        return strptime(version_month, "%b").tm_mon
                    except ValueError:  # pragma: no cover
                        raise ValueError(
                            f"Value {version_month} not recognised"
                            " as a month") from original
        else:  # pragma: no cover
            raise ValueError(
                f"Value {version_month} not recognised as a month")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

_WRITE_LOGS_TO_STDOUT = True


def _assert_logs_contains(level, log_records, submessage):
    for record in log_records:
        if record.levelname == level and submessage in record.getMessage():
            return
    if _WRITE_LOGS_TO_STDOUT:  # pragma: no cover
        for record in log_records:
            print(record)
    raise AssertionError(f"\"{submessage}\" not found in any {level} logs")


def _assert_logs_not_contains(level, log_records, submessage):
    for record in log_records:
        if _WRITE_LOGS_TO_STDOUT:  # pragma: no cover
            print(record)
        if record.levelname == level and submessage in record.getMessage():
            raise AssertionError(
                f"\"{submessage}\" found in any {level} logs")


def assert_logs_contains_once(level, log_records, message):
    """
    Checks if the log records contain exactly one record at the given
    level with the given sub-message.

    .. note::
        While this code does not depend on testfixtures,
        you will need testfixtures to generate the input data

    :param level: The log level. Probably "INFO", "WARNING" or "ERROR".
    :param log_records: list of log records returned by testfixtures.LogCapture
    :param submessage: String which should be part of a log record
    :raises AssertionError: If the submessage is not present in the log
    """
    found = False
    for record in log_records:
        if record.levelname == level and message == record.getMessage():
            if found:
                if _WRITE_LOGS_TO_STDOUT:  # pragma: no cover
                    for a_record in log_records:
                        print(a_record)
                raise AssertionError(
                    f"\"{message}\" found twice in  {level} logs")
            found = True
    if not found:
        if _WRITE_LOGS_TO_STDOUT:  # pragma: no cover
            for record in log_records:
                print(record)
        raise AssertionError(f"\"{message}\" not found in any {level} logs")


def assert_logs_error_contains(log_records, submessage):
    """
    Checks it the log records contain an ERROR log with this sub-message

    .. note::
        While this code does not depend on testfixtures,
        you will need testfixtures to generate the input data

    :param log_records: list of log records returned by testfixtures.LogCapture
    :param submessage: String which should be part of an ERROR log
    :raises AssertionError: If the submessage is not present in the log
    """
    _assert_logs_contains('ERROR', log_records, submessage)


def assert_logs_warning_contains(log_records, submessage):
    """
    Checks it the log records contain an WARNING log with this sub-message

    .. note::
        While this code does not depend on testfixtures,
        you will need testfixtures to generate the input data

    :param log_records: list of log records returned by testfixtures.LogCapture
    :param submessage: String which should be part of an WARNING log
    :raises AssertionError: If the submessage is not present in the log
    """
    _assert_logs_contains('WARNING', log_records, submessage)


def assert_logs_info_contains(log_records, sub_message):
    """
    Checks it the log records contain an INFO log with this sub-message

    .. note::
        While this code does not depend on testfixtures,
        you will need testfixtures to generate the input data

    :param log_records: list of log records returned by testfixtures.LogCapture
    :param sub_message: String which should be part of an INFO log
    :raises AssertionError: If the submessage is not present in the log
    """
    _assert_logs_contains('INFO', log_records, sub_message)


def assert_logs_error_not_contains(log_records, submessage):
    """
    Checks it the log records do not contain an ERROR log with this
    sub-message.

    .. note::
        While this code does not depend on testfixtures,
        you will need testfixtures to generate the input data

    :param log_records: list of log records returned by testfixtures.LogCapture
    :param submessage: String which should be part of an ERROR log
    :raises AssertionError: If the submessage is present in the log
    """
    _assert_logs_not_contains('ERROR', log_records, submessage)


def assert_logs_info_not_contains(log_records, submessage):
    """
    Checks it the log records do not contain an INFO log with this
    sub-message.

    .. note::
        While this code does not depend on testfixtures,
        you will need testfixtures to generate the input data

    :param log_records: list of log records returned by testfixtures.LogCapture
    :param submessage: String which should be part of an INFO log
    :raises AssertionError: If the submessage is present in the log
    """
    _assert_logs_not_contains('INFO', log_records, submessage)



# Copyright (c) 2018 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys


class IndexIsValue(object):
    """
    Tiny support class that implements ``object[x]`` by just returning
    ``x`` itself.

    Used for where you want a range from 1 to *N* but you don't know *N*.

    Clearly, operations that assume a finite list are *not* supported.
    """

    def __getitem__(self, key):
        return key

    def __len__(self):
        return sys.maxsize



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sqlite3
import sys
import time
from typing import Optional, Tuple
from spinn_utilities.abstract_context_manager import AbstractContextManager

_DDL_FILE = os.path.join(os.path.dirname(__file__), "db.sql")
_SECONDS_TO_MICRO_SECONDS_CONVERSION = 1000
DB_FILE_NAME = "logs.sqlite3"


def _timestamp():
    return int(time.time() * _SECONDS_TO_MICRO_SECONDS_CONVERSION)


class LogSqlLiteDatabase(AbstractContextManager):
    """
    Specific implementation of the Database for SQLite 3.

    .. note::
        **Not thread-safe on the same database.**
        Threads can access different DBs just fine.

    .. note::
        This totally relies on the way SQLite's type affinities function.
        You can't port to a different database engine without a lot of work.
    """

    __slots__ = [
        # the database holding the data to store
        "_db",
    ]

    def __init__(self, new_dict=False):
        """
        Connects to a log dict. The location of the file can be overridden
        using the ``C_LOGS_DICT`` environment variable.

        :param bool new_dict: Flag to say if this is a new dict or not.
            If True, clears and previous values.
            If False, makes sure the dict exists.
        """
        # To Avoid an Attribute error on close after an exception
        self._db = None
        database_file = self._database_file()
        if not new_dict:
            self._check_database_file(database_file)

        try:
            self._db = sqlite3.connect(database_file)
            self.__init_db()
            if new_dict:
                self.__clear_db()
        except Exception as ex:
            message = f"Error accessing c_logs_dict at {database_file}. "
            if 'C_LOGS_DICT' in os.environ:
                message += (
                    "This came from the environment variable 'C_LOGS_DICT'. ")
            else:
                message += (
                    "This is the default location. Set environment "
                    "variable 'C_LOGS_DICT' to use somewhere else.")
            if new_dict:
                message += "Check this is a location with write access."
            else:
                message += "Please rebuild the C code."
            raise FileNotFoundError(message) from ex

    def _database_file(self) -> str:
        """
        Finds the database file path.

        If environment variable C_LOGS_DICT exists that is used,
        otherwise the default path in this directory is used.

        :return: Absolute path to where the database file is or will be
        :rtype: str
        """
        if 'C_LOGS_DICT' in os.environ:
            return str(os.environ['C_LOGS_DICT'])

        script = sys.modules[self.__module__].__file__
        assert script is not None
        directory = os.path.dirname(script)
        return os.path.join(directory, DB_FILE_NAME)

    def _extra_database_error_message(self) -> str:
        """
        Adds a possible extra part to the error message.

        :return: A likely empty string
        :rtype: str
        """
        return ""

    def _check_database_file(self, database_file: str) -> None:
        """
        Checks the database file exists:

        :param str database_file: Absolute path to the database file
        :raises FileNotFoundErrorL If the file does not exists
        """
        if os.path.exists(database_file):
            return
        message = f"Unable to locate c_logs_dict at {database_file}. "
        if 'C_LOGS_DICT' in os.environ:
            message += (
                "This came from the environment variable 'C_LOGS_DICT'. ")
        message += self._extra_database_error_message()
        message += "Please rebuild the C code."
        raise FileNotFoundError(message)

    def __del__(self):
        self.close()

    def close(self):
        """
        Finalises and closes the database.
        """
        try:
            if self._db is not None:
                self._db.close()
        except Exception:  # pylint: disable=broad-except
            pass
        self._db = None

    def __init_db(self):
        """
        Set up the database if required.
        """
        self._db.row_factory = sqlite3.Row
        # Don't use memoryview / buffer as hard to deal with difference
        self._db.text_factory = str
        with open(_DDL_FILE, encoding="utf-8") as f:
            sql = f.read()
        self._db.executescript(sql)

    def __clear_db(self):
        with self._db:
            cursor = self._db.cursor()
            cursor.execute("DELETE FROM log")
            cursor.execute("UPDATE SQLITE_SEQUENCE SET SEQ=0 WHERE NAME='log'")
            cursor.execute("DELETE FROM file")
            cursor.execute(
                "UPDATE SQLITE_SEQUENCE SET SEQ=0 WHERE NAME='file'")
            cursor.execute("DELETE FROM directory")
            cursor.execute(
                "UPDATE SQLITE_SEQUENCE SET SEQ=0 WHERE NAME='directory'")

    def get_directory_id(self, src_path: str, dest_path: str) -> int:
        """
        gets the Ids for this directory. Making a new one if needed

        :param str src_path:
        :param str dest_path:
        :rtype: int
        """
        with self._db:
            cursor = self._db.cursor()
            # reuse the existing if it exists
            for row in self._db.execute(
                    """
                    SELECT directory_id
                    FROM directory
                    WHERE src_path = ? AND dest_path = ?
                    LIMIT 1
                    """, [src_path, dest_path]):
                return row["directory_id"]

            # create a new number
            cursor.execute(
                """
                INSERT INTO directory(src_path, dest_path)
                VALUES(?, ?)
                """, (src_path, dest_path))
            return cursor.lastrowid

    def get_file_id(self, directory_id: int, file_name: str) -> int:
        """
        Gets the id for this file, making a new one if needed.

        :param int directory_id:
        :param str file_name:
        :rtype: int
        """
        with self._db:
            # Make previous one as not last
            with self._db:
                cursor = self._db.cursor()
                cursor.execute(
                    """
                    UPDATE file SET last_build = 0
                    WHERE directory_id = ? AND file_name = ?
                    """, [directory_id, file_name])
                # always create new one to distinguish new from old logs
                cursor.execute(
                    """
                    INSERT INTO file(
                        directory_id, file_name, convert_time, last_build)
                    VALUES(?, ?, ?, 1)
                    """, (directory_id, file_name, _timestamp()))
                return cursor.lastrowid

    def set_log_info(
            self, log_level: int, line_num: int, original: str, file_id: int):
        """
        Saves the data needed to replace a short log back to the original.

        :param int log_level:
        :param int line_num:
        :param str original:
        :param int file_id:
        """
        with self._db:
            cursor = self._db.cursor()
            # reuse the existing number if nothing has changed
            cursor.execute(
                """
                UPDATE log SET
                    file_id = ?
                WHERE log_level = ? AND line_num = ? AND original = ?
                """, (file_id, log_level, line_num, original))

            if cursor.rowcount == 0:
                # create a new number if anything has changed
                cursor.execute(
                    """
                    INSERT INTO log(log_level, line_num, original, file_id)
                    VALUES(?, ?, ?, ?)
                    """, (log_level, line_num, original, file_id))
                return cursor.lastrowid
            else:
                for row in self._db.execute(
                        """
                        SELECT log_id
                        FROM log
                        WHERE log_level = ? AND line_num = ?
                            AND original = ? AND file_id = ?
                        LIMIT 1
                        """, (log_level, line_num, original, file_id)):
                    return row["log_id"]

    def get_log_info(self, log_id: str) -> Optional[Tuple[int, str, int, str]]:
        """
        Gets the data needed to replace a short log back to the original.

        :param str log_id: The int id as a String
        :rtype: tuple(int, str, int, str)
        """
        with self._db:
            for row in self._db.execute(
                    """
                    SELECT log_level, file_name, line_num , original
                    FROM current_file_view
                    WHERE log_id = ?
                    LIMIT 1
                    """, [log_id]):
                return (row["log_level"], row["file_name"], row["line_num"],
                        row["original"])
        return None

    def check_original(self, original: str):
        """
        Checks that an original log line has been added to the database.

        Mainly used for testing

        :param str original:
        :raises ValueError: If the original is not in the database
        """
        with self._db:
            for row in self._db.execute(
                    """
                    SELECT COUNT(log_id) as "counts"
                    FROM log
                    WHERE original = ?
                    """, ([original])):
                if row["counts"] == 0:
                    raise ValueError(f"{original} not found in database")

    def get_max_log_id(self):
        """
        Get the max id of any log message.

        :rtype: int
        """
        with self._db:
            for row in self._db.execute(
                    """
                    SELECT MAX(log_id) AS "max_id"
                    FROM log
                     """):
                return row["max_id"]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from spinn_utilities.config_holder import (
    get_config_int, get_config_int_or_none, get_config_str)


class SocketAddress(object):
    """
    Data holder for a socket interface for notification protocol.
    """
    __slots__ = [
        "_listen_port",
        "_notify_host_name",
        "_notify_port_no",
        "__hash"]

    def __init__(self, notify_host_name: Optional[str] = None,
                 notify_port_no: Optional[int] = None,
                 listen_port: Optional[int] = None):
        """
        :param notify_host_name:
            Host to talk to tell that the database (and application) is ready.
        :type notify_host_name: str or None
        :param notify_port_no:
            Port to talk to tell that the database (and application) is ready.
        :type notify_port_no: int or None
        :param listen_port:
            Port on which to listen for an acknowledgement that the
            simulation should start.
        :type listen_port: int or None
        """
        if notify_port_no is None:
            notify_port_no = get_config_int("Database", "notify_port")
        else:
            notify_port_no = int(notify_port_no)
        if notify_host_name is None:
            notify_host_name = get_config_str("Database", "notify_hostname")
        elif notify_host_name == "0.0.0.0":
            notify_host_name = "localhost"
        else:
            notify_host_name = str(notify_host_name)
        if listen_port is None:
            listen_port = get_config_int_or_none("Database", "listen_port")
        else:
            listen_port = int(listen_port)
        self._notify_host_name = notify_host_name
        self._notify_port_no = notify_port_no
        self._listen_port = listen_port
        self.__hash: Optional[int] = None

    @property
    def notify_host_name(self) -> str:
        """
        The host name or IP address to send a notification message to.
        """
        return self._notify_host_name

    @property
    def notify_port_no(self) -> int:
        """
        The UDP port number to send a notification message to.
        """
        return self._notify_port_no

    @property
    def listen_port(self) -> Optional[int]:
        """
        The port to listen to for responses.
        """
        return self._listen_port

    def __eq__(self, other):
        if not isinstance(other, SocketAddress):
            return False
        return (self._notify_host_name == other.notify_host_name and
                self._notify_port_no == other.notify_port_no and
                self._listen_port == other.listen_port)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        if self.__hash is None:
            self.__hash = hash((self._listen_port, self._notify_host_name,
                                self._notify_port_no))
        return self.__hash

    def __repr__(self):
        return (f"SocketAddress({repr(self._notify_host_name)}, "
                f"{self._notify_port_no}, {self._listen_port})")



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Type
from spinn_utilities.exceptions import (
    DataNotMocked, DataNotYetAvialable, NotSetupException, ShutdownException,
    SpiNNUtilsException)


class DataStatus(Enum):
    """
    Different states the Data can be in.

    This class is designed to used internally by UtilsDataView
    """
    #: No setup calls have been done yet.
    NOT_SETUP = (0, NotSetupException)
    #: The system is (to be) setup in mocked mode for unit testing.
    MOCKED = (1, DataNotMocked)
    #: The system is (to be) setup for running user code.
    SETUP = (2, DataNotYetAvialable)
    #: The system has been shut down.
    SHUTDOWN = (3, ShutdownException)

    def __new__(cls, *args) -> 'DataStatus':
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    def __init__(self, value, exception: Type[SpiNNUtilsException]):
        # pylint: disable=unused-argument
        self._exception = exception

    def exception(self, data) -> SpiNNUtilsException:
        """
        Returns an instance of the most suitable data-not-available exception.

        :param data: Parameter to pass to the relevant constructor.
        :rtype: ~spinn_utilities.exceptions.SpiNNUtilsException
        """
        return self._exception(data)



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class SpiNNUtilsException(Exception):
    """
    Superclass of all exceptions from the SpiNNUtils module.
    """


class UnexpectedStateChange(SpiNNUtilsException):
    """
    Raised when trying to change the state in an unexpected way.
    """


class NotSetupException(SpiNNUtilsException):
    """
    Raised when trying to get data before simulator has been setup.
    """

    def __init__(self, data):
        super().__init__(f"Requesting {data} is not valid before setup")


class InvalidDirectory(SpiNNUtilsException):
    """
    Raised when trying to set an invalid directory.
    """
    def __init__(self, name, value):
        super().__init__(f"Unable to set {name} has {value} is not a dir.")


class DataNotYetAvialable(SpiNNUtilsException):
    """
    Raised when trying to get data before simulator has created it.
    """
    def __init__(self, data):
        super().__init__(f"{data} has not yet been created.")


class DataNotMocked(DataNotYetAvialable):
    """
    Raised when trying to get data before a mocked simulator has created it.
    """
    def __init__(self, data):
        super().__init__(f"MOCK {data}")


class ShutdownException(SpiNNUtilsException):
    """
    Raised when trying to get simulator data after it has been shut down.
    """
    def __init__(self, data):
        super().__init__(f"Requesting {data} is not valid after end")


class IllegalWriterException(SpiNNUtilsException):
    """
    Raised when trying to create a writer other than setup or Mock.
    """


class SimulatorNotSetupException(SpiNNUtilsException):
    """
    Raised when trying to get simulator before it has been setup.
    """


class SimulatorShutdownException(SpiNNUtilsException):
    """
    Raised when trying to get simulator after it has been shut down.
    """


class SimulatorRunningException(SpiNNUtilsException):
    """
    Raised when trying an action that should not happen while the simulator
    is running.
    """


class SimulatorNotRunException(SpiNNUtilsException):
    """
    Raised when trying to reset or stop before starting.
    """


class ConfigException(SpiNNUtilsException):
    """
    Raised when reading or setting configurations went wrong.
    """


class UnexpectedCException(SpiNNUtilsException):
    """
    Raised when the converter (which replaces log messages) found a pattern
    it did not expect and cannot handle.
    """



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import List, Optional
from spinn_utilities.ordered_set import OrderedSet


class ExecutableFinder(object):
    """
    Manages a set of folders in which to search for binaries,
    and allows for binaries to be discovered within this path.
    """
    __slots__ = [
        "_binary_search_paths",
        "_binary_log",
        "_paths_log"]

    def __init__(self) -> None:
        global_reports = os.environ.get("GLOBAL_REPORTS", None)
        if global_reports:
            if not os.path.exists(global_reports):
                # It might now exist if run in parallel
                try:
                    os.makedirs(global_reports)
                except Exception:  # pylint: disable=broad-except
                    pass
            self._paths_log: Optional[str] = os.path.join(
                global_reports, "binary_paths_used.log")
            self._binary_log: Optional[str] = os.path.join(
                global_reports, "binary_files_used.log")

        else:
            self._paths_log = None
            self._binary_log = None

        self._binary_search_paths: OrderedSet[str] = OrderedSet()

    def add_path(self, path: str) -> None:
        """
        Adds a path to the set of folders to be searched.  The path is
        added to the end of the list, so it is searched after all the
        paths currently in the list.

        :param str path: The path to add
        """
        self._binary_search_paths.add(path)
        if self._paths_log:
            try:
                with open(self._paths_log, "a", encoding="utf-8") as log_file:
                    log_file.write(path)
                    log_file.write("\n")
            except Exception:  # pylint: disable=broad-except
                pass

    @property
    def binary_paths(self) -> str:
        """
        The set of folders to search for binaries, as a printable
        colon-separated string.

        :rtype: str
        """
        return " : ".join(self._binary_search_paths)

    def get_executable_path(self, executable_name: str) -> str:
        """
        Finds an executable within the set of folders. The set of folders
        is searched sequentially and the first match is returned.

        :param str executable_name: The name of the executable to find
        :return: The full path of the discovered executable
        :rtype: str
        :raises KeyError: If no executable was found in the set of folders
        """
        # Loop through search paths
        for path in self._binary_search_paths:
            # Rebuild filename
            potential_filename = os.path.join(path, executable_name)

            # If this filename exists, return it
            if os.path.isfile(potential_filename):
                if self._binary_log:
                    try:
                        with open(self._binary_log, "a", encoding="utf-8") \
                                as log_file:
                            log_file.write(potential_filename)
                            log_file.write("\n")
                    except Exception:  # pylint: disable=broad-except
                        pass
                return potential_filename

        # No executable found
        raise KeyError(f"Executable {executable_name} not found in paths "
                       f"f{list(self._binary_search_paths)}")

    def get_executable_paths(self, executable_names: str) -> List[str]:
        """
        Finds each executables within the set of folders.

        The names are assumed to be comma separated
        The set of folders is searched sequentially
        and the first match for each name is returned.

        Names not found are ignored and not added to the list.

        :param str executable_names: The name of the executable to find.
            Assumed to be comma separated.
        :return:
            The full path of the discovered executable, or ``None`` if no
            executable was found in the set of folders
        :rtype: list(str)
        """
        results = list()
        for name in executable_names.split(","):
            try:
                results.append(self.get_executable_path(name))
            except KeyError:
                pass
        return results

    def check_logs(self) -> None:
        """
        Compares the aplx files used against the ones available for use

        Reports the aplx files never used.
        """
        if not self._paths_log:
            print("environ GLOBAL_REPORTS not set!")
            return

        folders = set()
        with open(self._paths_log, "r", encoding="utf-8") as log_file:
            for line in log_file:
                folders.add(line.strip())

        in_folders = set()
        for folder in folders:
            try:
                for file_name in os.listdir(folder):
                    if file_name.endswith(".aplx"):
                        in_folders.add(os.path.join(folder, file_name))
            except Exception:  # pylint: disable=broad-except
                # Skip folders not found
                pass

        used_binaries = set()
        if self._binary_log:
            with open(self._binary_log, "r", encoding="utf-8") as log_file:
                for line in log_file:
                    used_binaries.add(line.strip())

        missing = in_folders - used_binaries
        print(f"{len(used_binaries)} binaries asked for. "
              f"{len(missing)} binaries never asked for.")
        if len(missing) > 0:
            print("Binaries asked for are:")
            for binary in (used_binaries):
                print(binary)
            print("Binaries never asked for are:")
            for binary in (missing):
                print(binary)

    def clear_logs(self) -> None:
        """
        Deletes log files from previous runs
        """
        if not self._paths_log:
            print("environ GLOBAL_REPORTS not set!")
            return
        if self._paths_log and os.path.isfile(self._paths_log):
            os.remove(self._paths_log)
        if self._binary_log and os.path.isfile(self._binary_log):
            os.remove(self._binary_log)


if __name__ == "__main__":
    ef = ExecutableFinder()
    try:
        ef.check_logs()
        ef.clear_logs()
    except Exception as ex:  # pylint: disable=broad-except
        print(ex)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from collections.abc import MutableSet
from typing import (
    Any, Dict, Iterable, Iterator, Optional, Generic, TypeVar)

#: :meta private:
T = TypeVar("T")


class OrderedSet(MutableSet, Generic[T]):
    """
    A set like Object where peek, pop and iterate are in insert order
    """
    __slots__ = (
        "_map",
    )

    def __init__(self, iterable: Optional[Iterable[T]] = None):
        # pylint: disable=super-init-not-called
        # Always use OrderedDict as plain dict does not support
        # __reversed__ and key indexing
        self._map: Dict[T, None] = OrderedDict()

        # or is overridden in mutable set; calls add on each element
        if iterable is not None:
            self.update(iterable)

    def add(self, value: T):
        if value not in self._map:
            self._map[value] = None

    def discard(self, value: T):
        if value in self._map:
            self._map.pop(value)

    def __iter__(self) -> Iterator[T]:
        return self._map.__iter__()

    def __reversed__(self) -> Iterator[T]:
        return self._map.__reversed__()

    def peek(self, last: bool = True) -> T:
        """
        Retrieve the first element from the set without removing it
        :param last:
        :return:
        """
        if not self._map:  # i.e., is self._map empty?
            raise KeyError('set is empty')
        if last:
            return next(reversed(self))
        else:
            return next(iter(self))

    def __len__(self) -> int:
        return len(self._map)

    def __contains__(self, key: Any) -> bool:
        return key in self._map

    def update(self, iterable: Iterable[T]):
        """
        Updates the set by adding each item in order

        :param iterable:
        """
        for item in iterable:
            self.add(item)

    def pop(self, last: bool = True) -> T:  # pylint: disable=arguments-differ
        key = self.peek(last)
        self.discard(key)
        return key

    def __repr__(self) -> str:
        if not self._map:  # i.e., is self._map empty?
            return f'{self.__class__.__name__}()'
        return f'{self.__class__.__name__}({list(self)})'

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and self._map == other._map
        return set(self) == set(other)

    def __ne__(self, other: Any) -> bool:
        """
        Comparison method for comparing ordered sets.

        :param other: instance of OrderedSet
        :rtype: None
        """
        return not self.__eq__(other)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import traceback
from spinn_utilities.overrides import overrides


def all_modules(directory, prefix, remove_pyc_files=False):
    """
    List all the python files found in this directory giving then the prefix.

    Any file that ends in either ``.py`` or ``.pyc`` is assume a python module
    and added to the result set.

    :param str directory: path to check for python files
    :param str prefix: package prefix top add to the file name
    :return: set of python package names
    :rtype: set(str)
    """
    results = set()
    for module in os.listdir(directory):
        if module == "__init__.py":
            results.add(prefix)
        elif module == "__init__.pyc":
            results.add(prefix)
            if remove_pyc_files:  # pragma: no cover
                full_path = os.path.join(directory, module)
                print("Deleting: " + full_path)
                os.remove(full_path)
        elif module[-3:] == ".py":
            results.add(prefix + "." + module[:-3])
        elif module[-4:] == ".pyc":
            results.add(prefix + "." + module[:-4])
            if remove_pyc_files:  # pragma: no cover
                full_path = os.path.join(directory, module)
                print("Deleting: " + full_path)
                os.remove(full_path)
        elif module != "__pycache__":
            full_path = os.path.join(directory, module)
            if os.path.isdir(full_path):
                results.update(all_modules(
                    full_path, prefix + "." + module, remove_pyc_files))
    return results


def load_modules(
        directory, prefix, remove_pyc_files=False, exclusions=None,
        gather_errors=True):
    """
    Loads all the python files found in this directory, giving them the
    specified prefix.

    Any file that ends in either ``.py`` or ``.pyc`` is assume a python module
    and added to the result set.

    :param str directory: path to check for python files
    :param str prefix: package prefix top add to the file name
    :param bool remove_pyc_files: True if ``.pyc`` files should be deleted
    :param list(str) exclusions: a list of modules to exclude
    :param bool gather_errors:
        True if errors should be gathered, False to report on first error
    :return: None
    """
    if exclusions is None:
        exclusions = []
    modules = all_modules(directory, prefix, remove_pyc_files)
    errors = list()
    for module in modules:
        if module in exclusions:
            print("SKIPPING " + module)
            continue
        try:
            __import__(module)
        except Exception:  # pylint: disable=broad-except
            print(f"Error with {module}")
            if gather_errors:
                errors.append((module, sys.exc_info()))
            else:
                raise

    for module, (exc_type, exc_value, exc_traceback) in errors:
        print(f"Error importing {module}:")
        for line in traceback.format_exception(
                exc_type, exc_value, exc_traceback):
            for line_line in line.split("\n"):
                if line_line:
                    print("  ", line_line.rstrip())
    if errors:
        raise ImportError(f"Error when importing, starting at {prefix}")


def load_module(
        name, remove_pyc_files=False, exclusions=None, gather_errors=True):
    """
    Loads this modules and all its children.

    :param str name: name of the modules
    :param bool remove_pyc_files: True if ``.pyc`` files should be deleted
    :param list(str) exclusions: a list of modules to exclude
    :param bool gather_errors:
        True if errors should be gathered, False to report on first error
    :return: None
    """
    overrides.check_types()
    if exclusions is None:
        exclusions = []
    module = __import__(name)
    path = module.__file__
    directory = os.path.dirname(path)
    load_modules(directory, name, remove_pyc_files, exclusions, gather_errors)


if __name__ == '__main__':  # pragma: no cover
    load_module("spinn_utilities", True)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import (
    Dict, Generic, Iterable, Iterator, Optional, Sequence, Tuple, overload,
    TYPE_CHECKING, Union)
from spinn_utilities.overrides import overrides
from .abstract_dict import AbstractDict, T, _StrSeq, _Keys
from .abstract_view import AbstractView
if TYPE_CHECKING:
    from .range_dictionary import RangeDictionary
    from .ranged_list import _ValueType


class _SliceView(AbstractView[T], Generic[T]):
    __slots__ = ("_start", "_stop")

    def __init__(self, range_dict: RangeDictionary, start: int, stop: int):
        """
        Use :py:meth:`RangeDictionary.view_factory` to create views
        """
        super().__init__(range_dict)
        self._start = start
        self._stop = stop

    def __str__(self) -> str:
        return f"View with range: {self._start} to {self._stop}"

    @overrides(AbstractDict.ids)
    def ids(self) -> Sequence[int]:
        return range(self._start, self._stop)

    @overload
    def get_value(self, key: str) -> T:
        ...

    @overload
    def get_value(self, key: Optional[_StrSeq]) -> Dict[str, T]:
        ...

    @overrides(AbstractDict.get_value)
    def get_value(self, key: _Keys) -> Union[T, Dict[str, T]]:
        if isinstance(key, str):
            return self._range_dict.get_list(key).get_single_value_by_slice(
                slice_start=self._start, slice_stop=self._stop)
        elif key is None:
            return {
                k: self._range_dict.get_list(k).get_single_value_by_slice(
                    slice_start=self._start, slice_stop=self._stop)
                for k in self._range_dict.keys()}
        else:
            return {
                k: self._range_dict.get_list(k).get_single_value_by_slice(
                    slice_start=self._start, slice_stop=self._stop)
                for k in key}

    def update_safe_iter_all_values(self, key: str) -> Iterable[T]:
        """
        Iterate over the Values in a way that will work even between updates

        :param str key:
        :rtype: iter
        """
        ranged_list = self._range_dict.get_list(key)
        for the_id in self.ids():
            yield ranged_list.get_value_by_id(the_id=the_id)

    @overload
    def iter_all_values(
            self, key: str, update_safe=False) -> Iterator[T]:
        ...

    @overload
    def iter_all_values(
            self, key: Optional[_StrSeq] = None,
            update_safe=False) -> Iterator[Dict[str, T]]:
        ...

    @overrides(AbstractDict.iter_all_values, extend_defaults=True)
    def iter_all_values(self, key: _Keys = None, update_safe: bool = False):
        if isinstance(key, str):
            if update_safe:
                return self.update_safe_iter_all_values(key)
            return self._range_dict.get_list(key).iter_by_slice(
                slice_start=self._start, slice_stop=self._stop)
        return self._range_dict.iter_values_by_slice(
            key=key, slice_start=self._start, slice_stop=self._stop,
            update_safe=update_safe)

    @overrides(AbstractDict.set_value)
    def set_value(self, key: str, value: _ValueType,
                  use_list_as_value: bool = False):
        self._range_dict.get_list(key).set_value_by_slice(
            slice_start=self._start, slice_stop=self._stop, value=value,
            use_list_as_value=use_list_as_value)

    @overload
    def iter_ranges(self, key: str) -> Iterator[Tuple[int, int, T]]:
        ...

    @overload
    def iter_ranges(self, key: Optional[_StrSeq] = None) -> Iterator[
            Tuple[int, int, Dict[str, T]]]:
        ...

    @overrides(AbstractDict.iter_ranges)
    def iter_ranges(self, key: _Keys = None
                    ) -> Union[Iterator[Tuple[int, int, T]],
                               Iterator[Tuple[int, int, Dict[str, T]]]]:
        return self._range_dict.iter_ranges_by_slice(
            key=key, slice_start=self._start, slice_stop=self._stop)



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable
_Check = Callable[[int], bool]


def find_max_success(max_possible: int, check: _Check) -> int:
    """
    Finds the maximum value that will pass the check

    :param max_possible: The maximum value that should be tested.
    :type max_possible: int
    :param check: A Boolean function that given an int value returns
        True for every value up and including the cut-off and
        False for every value greater than the cut-off
    :type check: ~typing.Callable[[int], bool]
    :return: The highest value that returns true for the check
        but is not more than the max_possible
    """
    # Check the max to avoid multiple steps if that works
    if check(max_possible):
        return max_possible
    return search_for_max_success(0, max_possible, check)


def search_for_max_success(
        best_success: int, min_fail: int, check: _Check) -> int:
    """
    Finds the maximum value in the range that pass the check

    :param best_success: A minimum value that needs not be tested because it
        is either known to succeed or is a flag for total failure.
        Can be negative
    :type best_success: int
    :param min_fail: A maximum value that needs not be tested because it
        is either known to fail or one more than the maximum interesting value
        but must be greater than best_success but may also be negative
    :type min_fail: int
    :param check: A Boolean function that given an int value returns
        True for every value up and including the cut-off and
        False for every value greater than the cut-off
    :type check: ~typing.Callable[[int], bool]
    :return: The highest value that returns true in the range between
        `best_success` and `min_fail` (both exclusive ends) or `best_success`
        if the whole range fails or is empty.
    """
    # Check if there are still values in the middle to check
    if best_success >= min_fail - 1:
        return best_success
    # Find the middle
    mid_point = (best_success + min_fail) // 2
    # Check the middle
    if check(mid_point):
        # Best it the middle or more
        return search_for_max_success(mid_point, min_fail, check)
    else:
        # Best it the less than the middle
        return search_for_max_success(best_success, mid_point, check)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import reduce
import logging
import math
from typing import Any, Sequence, Union, overload
from typing_extensions import TypeGuard
from spinn_utilities.log import FormatAdapter


logger = FormatAdapter(logging.getLogger(__name__))
FINISHED_FILENAME = "finished"


def is_singleton(value: Any) -> TypeGuard[Union[bool, int, float]]:
    """
    Tests whether the value is a singleton.

    Singleton types are strings and any other class that can not be
    iterated.

    Strings are considered singleton as only rarely will someone use a string
    to represent an iterable of characters.
    """
    return not hasattr(value, '__iter__') or isinstance(value, str)


def _lcm(a: int, b: int) -> int:
    return (a * b) // math.gcd(a, b)


# A sequence of integers..
@overload
def lcm(numbers: Sequence[int], /) -> int:
    ...


# ... or several integers. All in positional-only parameters.
@overload
def lcm(number: int, /, *numbers: int) -> int:
    ...


def lcm(*numbers) -> int:
    """
    Lowest common multiple of 0, 1 or more integers.

    GIGO: If any of the values are anything except positive `int` values
    this function will either produce incorrect results or raise an exception.

    :param numbers:
        The Positive integers to get the Lowest common multiple for.
        This can be zero, one or more int values or
        a singleton which is an iterator (possibly empty) of `int`\\s.
    :return: the  Lowest common multiple,
        or 1 if `numbers` is empty or an empty iterator
    :rtype: int
    :raises TypeError: If any value cannot be interpreted as an integer
    :raises ZeroDivisionError: May be raised if one of the values is zero
    """
    if len(numbers) == 1:
        try:
            return reduce(_lcm, iter(numbers[0]), 1)
        except TypeError:
            return numbers[0]
    return reduce(_lcm, numbers, 1)


# A sequence of integers..
@overload
def gcd(numbers: Sequence[int], /) -> int:
    ...


# ... or several integers. All in positional-only parameters.
@overload
def gcd(number: int, /, *numbers: int) -> int:
    ...


def gcd(*numbers) -> int:
    """
    Greatest common divisor of 1 or more integers.

    GIGO: If any of the values are anything except positive `int` values
    this function will either produce incorrect results or raise an exception.

    :param numbers: The positive integers to get the GCD for.
        This can be one or more `int` values or
        a singleton which is an iterator (*not* empty) of `int`\\s.
    :return: the GCD
    :rtype: int
    :raises TypeError: If any value cannot be interpreted as an integer or
        if no values are provided
    :raises ZeroDivisionError: May be raised if one of the values is zero
    """
    if len(numbers) == 1:
        try:
            return reduce(math.gcd, iter(numbers[0]))
        except TypeError:
            return numbers[0]
    return reduce(math.gcd, numbers)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import (
    Dict, FrozenSet, Iterable, Iterator, MutableSequence, Optional, Sequence,
    Set, Tuple, Union,
    Generic, TypeVar, overload)
from typing_extensions import TypeAlias
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
#: :meta private:
T = TypeVar("T")
# Can't be Iterable[str] or Sequence[str] because that includes str itself
_StrSeq: TypeAlias = Union[
    MutableSequence[str], Tuple[str, ...], FrozenSet[str], Set[str]]
_Keys: TypeAlias = Optional[Union[str, _StrSeq]]


class AbstractDict(Generic[T], metaclass=AbstractBase):
    """
    Base class for the :py:class:`RangeDictionary` and *all* views.
    This allows the users to not have to worry if they have a view.
    """
    __slots__ = ()

    @overload
    def get_value(self, key: str) -> T:
        ...

    @overload
    def get_value(self, key: Optional[_StrSeq]) -> Dict[str, T]:
        ...

    @abstractmethod
    def get_value(self, key: _Keys) -> Union[T, Dict[str, T]]:
        """
        Gets a single shared value for all IDs covered by this view.

        :param key: The key or keys to get the value of. Use `None` for all
        :type key: str or iterable(str) or None
        :return: If key is a str, this returns the single object.
            If key is iterable (list, tuple, set, etc.) of str (or `None`),
            returns a dictionary object
        :raises ~spinn_utilities.ranged.MultipleValuesException:
            If even one of the keys has multiple values set.
            But not if other keys not asked for have multiple values
        """
        raise NotImplementedError

    @abstractmethod
    def keys(self) -> Iterable[str]:
        """
        Returns the keys in the dictionary

        :return: keys in the dict
        """
        raise NotImplementedError

    @abstractmethod
    def set_value(
            self, key: str, value: T, use_list_as_value: bool = False):
        """
        Sets a already existing key to the new value.
        All IDs in the whole range or view will have this key set.

        .. warning::
            This method does not allow adding keys.
            Using `dict[str] =` will add a new key, but it is not supported
            for views.

        .. warning::
            If a View is created over multiple ranges this method would
            raise a `KeyError` if any the ranges does not have the key.
            (Currently multiple ranges not yet supported.)

        :param key: key to value being set
        :type key: str
        :param value: any object
        :param use_list_as_value: True if the value *is* a list
        :raise KeyError: If a new key is being used.
        """
        raise NotImplementedError

    @abstractmethod
    def ids(self) -> Sequence[int]:
        """
        Returns the IDs in range or view.
        If the view is setup with IDs out of numerical order the order used
        to create the view is maintained.

        .. note::
            If indexing into a view, you are picking the Nth ID.
            So if the IDs are `[2,3,4,5]` the `view[2]` will be the data for
            ID `4` and not `2`

        :return: list of IDs
        :rtype: list(int)
        """
        raise NotImplementedError

    @overload
    def iter_all_values(self, key: str, update_safe=False) -> Iterator[T]:
        ...

    @overload
    def iter_all_values(self, key: Optional[_StrSeq],
                        update_safe: bool = False) -> Iterator[Dict[str, T]]:
        ...

    @abstractmethod
    def iter_all_values(self, key: _Keys, update_safe: bool = False):
        """
        Iterates over the value(s) for all IDs covered by this view.
        There will be one yield for each ID even if values are repeated.

        :param key:
            The key or keys to get the value of. Use `None` for all keys
        :type key: str or iterable(str) or None
        :param update_safe: If set True the iteration will work even if values
            are updated during iteration. If left False the iterator may be
            faster but behaviour is *undefined* and *unchecked* if *any*
            values are changed during iteration.
        :return: If key is a str, this yields single objects.
            If key is iterable (list, tuple, set, etc.) of str (or `None`),
            yields dictionary objects
        """
        raise NotImplementedError

    @overload
    def get_ranges(self, key: None = None) -> Sequence[Tuple[
            int, int, Dict[str, T]]]:
        ...

    @overload
    def get_ranges(self, key: str) -> Sequence[Tuple[int, int, T]]:
        ...

    @overload
    def get_ranges(self, key: _StrSeq) -> Sequence[Tuple[
            int, int, Dict[str, T]]]:
        ...

    def get_ranges(self, key: _Keys = None) -> Sequence[Tuple[
            int, int, Union[T, Dict[str, T]]]]:
        """
        Lists the ranges(s) for all IDs covered by this view.
        There will be one yield for each range which may cover one or
        more IDs.

        .. note::
            As the data is created in a single call this is not affected
            by any updates.

        :param key: The key or keys to get the value of. Use `None` for all
        :type key: str or iterable(str) or None
        :return: List of tuples of (`start`, `stop`, `value`).
            `start` is *inclusive* so is the first ID in the range.
            `stop` is *exclusive* so is the last ID in the range + 1.
            If `key` is a str, `value` is a single object.
            If `key` is iterable (list, tuple, set, etc.) of str (or `None`)
            `value` is a dictionary object
        """
        return list(self.iter_ranges(key=key))

    @overload
    def iter_ranges(self, key: str) -> Iterator[Tuple[int, int, T]]:
        ...

    @overload
    def iter_ranges(self, key: Optional[_StrSeq]) -> Iterator[Tuple[
            int, int, Dict[str, T]]]:
        ...

    @abstractmethod
    def iter_ranges(self, key: _Keys = None
                    ) -> Union[Iterator[Tuple[int, int, T]],
                               Iterator[Tuple[int, int, Dict[str, T]]]]:
        """
        Iterates over the ranges(s) for all IDs covered by this view.
        There will be one yield for each range which may cover one or
        more IDs.

        .. warning::
            This iterator is *not* update safe!
            Behaviour is *undefined* and *unchecked* if *any* values are
            changed during iteration.

        :param key: The key or keys to get the value of. Use `None` for all
        :type key: str or iterable(str) or None
        :return: yields tuples of (`start`, `stop`, `value`).
            `start` is *inclusive* so is the first ID in the range.
            `stop` is *exclusive* so is the last ID in the range + 1.
            If `key` is a str, `value` is a single object.
            If `key` is iterable (list, tuple, set, etc.) of str (or `None`),
            `value` is a dictionary object
        """
        raise NotImplementedError

    @abstractmethod
    def get_default(self, key: str) -> Optional[T]:
        """
        Gets the default value for a single key.
        Unless changed, the default is the original value.

        .. note::
            Does not change any values but only changes what ``reset_value``
            would do

        :param key: Existing dict key
        :type key: str
        :return: default for this key.
        """
        raise NotImplementedError

    def items(self) -> Sequence[Tuple[str, T]]:
        """
        Returns a list of (``key``, ``value``) tuples.
        Works only if the whole ranges/view has single values.

        If the ``key`` is a str, the ``value``\\s are single objects.
        If the ``key`` is iterable (list, tuple, set, etc.) of str (or `None`),
        the ``value``\\s are dictionary objects.

        :return: List of (``key``, ``value``) tuples
        :rtype: list(tuple)
        :raises ~spinn_utilities.ranged.MultipleValuesException:
            If even one of the keys has multiple values set.
        """
        results = []
        for key in self.keys():
            value = self.get_value(key)
            results.append((key, value))
        return results

    def iteritems(self) -> Iterator[Tuple[str, T]]:
        """
        Iterates over the (``key``, ``value``) tuples.
        Works only if the whole ranges/view has single values.

        If the ``key`` is a str, the ``value``\\s are single objects.
        If the ``key`` is iterable (list, tuple, set, etc.) of str (or `None`),
        the ``value``\\s are dictionary objects

        This function is safe for value updates but may miss new keys
        added during iteration.

        :return: yield (``key``, ``value``) tuples
        :rtype: iterable(tuple)
        :raises ~spinn_utilities.ranged.MultipleValuesException:
            If even one of the keys has multiple values set.
        """
        for key in self.keys():
            yield (key, self.get_value(key))

    def values(self) -> Sequence[T]:
        """
        Returns a list of values.
        Works only if the whole ranges/view has single values.

        If a key is a str, its value is a single object.
        If a key is an iterable (list, tuple, set, etc.) of str (or `None`),
        its value is a dictionary object.

        :return: List of values
        :raises ~spinn_utilities.ranged.MultipleValuesException:
            If even one of the keys has multiple values set.
        """
        results = []
        for key in self.keys():
            value = self.get_value(key)
            results.append(value)
        return results

    def itervalues(self) -> Iterator[T]:
        """
        Iterates over the values.
        Works only if the whole ranges/view has single values.

        If a key is a str, its value is a single object.
        If a key is an iterable (list, tuple, set, etc.) of str (or `None`),
        its value is a dictionary object.

        This function is safe for value updates but may miss new keys
        added during iteration.

        :return: yield values
        :raises ~spinn_utilities.ranged.MultipleValuesException:
            If even one of the keys has multiple values set.
        """
        for key in self.keys():
            yield self.get_value(key)

    def __contains__(self, key: Union[str, int]) -> bool:
        """
        Checks if the key is a dictionary key or a range ID.

        :param key: Dictionary key or ID to check
        :type key: str or int
        :return: True if the str key is one of the dict keys or
            if the int key is one of the range IDs. Otherwise False
        :rtype: bool
        """
        if isinstance(key, str):
            return key in self.keys()
        if isinstance(key, int):
            return key in self.ids()
        raise KeyError(f"Unexpected key type: {type(key)}")

    def has_key(self, key: str) -> bool:
        """
        As the Deprecated dict ``has_keys`` function.

        .. note::
            Int keys to IDs are not supported.

        :param key: the key
        :type key: str
        :return: If the key is in dict
        :rtype: bool
        """
        return key in self.keys()

    def reset(self, key: str):
        """
        Sets the value(s) for a single key back to the default value.

        :param key: Existing dict key
        :type key: str
        :param default: Value to be used by reset
        """
        default = self.get_default(key)
        if default is None:
            raise ValueError(f"key '{key}' is not resettable")
        self.set_value(key, default)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sized
import itertools
import logging
import sys
from typing import Any, Optional, Sequence, SupportsInt, Tuple, Union
from typing_extensions import TypeAlias, TypeGuard
import numpy

logger = logging.getLogger(__file__)

#: Type of integers for selectors
_Integer: TypeAlias = Union[int, numpy.integer, SupportsInt]
#: The type of selectors; properly, SupportsInt should exclude numpy.ndarray
#: because that's there just to handle special edge cases.
Selector: TypeAlias = Union[
    None, _Integer, slice, Sequence[Union[bool, numpy.bool_]],
    Sequence[_Integer]]


def _is_iterable_selector(selector: Selector) -> TypeGuard[
        Union[Sequence[Union[bool, numpy.bool_]], Sequence[_Integer]]]:
    # Check selector is an iterable using pythonic try
    try:
        iterator = iter(selector)  # type: ignore[arg-type]
    except TypeError:
        return False
    return iterator is not None


class AbstractSized(object):
    """
    Base class for slice and ID checking against size.
    Subclasses of this support the `len()` built-in.
    """

    __slots__ = (
        "_size", )

    def __init__(self, size: Union[int, float]):
        """
        :param int size: Fixed length of the list.
        """
        # Strictly doesn't need to be int, but really should be!
        self._size = max(int(round(size)), 0)

    def __len__(self) -> int:
        """
        Size of the list, irrespective of actual values.

        :return: the initial and Fixed size of the list
        """
        return self._size

    @staticmethod
    def _is_id_type(the_id: Any) -> TypeGuard[Union[int, SupportsInt]]:
        """
        Check if the given ID has a type acceptable for IDs.
        """
        return isinstance(the_id, (int, SupportsInt)) and (
            not isinstance(the_id, (float, numpy.ndarray)))

    def _check_id_in_range(self, the_id: Union[int, SupportsInt]) -> int:
        if not self._is_id_type(the_id):
            raise TypeError(f"Invalid argument type {type(the_id)}.")
        the_id = int(the_id)
        if not 0 <= the_id < self._size:
            raise IndexError(f"The index {the_id} is out of range.")
        return the_id

    def _check_slice_in_range(
            self, slice_start: Optional[int],
            slice_stop: Optional[int]) -> Tuple[int, int]:
        # Fix types
        if slice_start is None:
            slice_start = 0
        elif not self._is_id_type(slice_start):
            raise TypeError(f"Invalid argument type {type(slice_start)}.")
        else:
            slice_start = int(slice_start)
        if slice_stop is None or slice_stop == sys.maxsize:
            slice_stop = self._size
        elif not self._is_id_type(slice_stop):
            raise TypeError(f"Invalid argument type {type(slice_stop)}.")
        else:
            slice_stop = int(slice_stop)

        if slice_start < 0:
            slice_start = self._size + slice_start
            if slice_start < 0:
                logger.warning(
                    "Specified slice start was %d while size is only %d. "
                    "Therefore slice will start at index 0",
                    int(slice_start) - self._size, self._size)
                slice_start = 0
        elif slice_start >= len(self):
            logger.warning(
                "Specified slice start was %d while size is only %d. "
                "Therefore slice will be empty",
                slice_start - self._size, self._size)
            return (self._size, self._size)

        if slice_stop < 0:
            slice_stop = self._size + slice_stop

        if slice_start > slice_stop:
            logger.warning(
                "Specified slice has a start %d greater than its stop %d "
                "(based on size %d). Therefore slice will be empty",
                slice_start, slice_stop, self._size)
            return (self._size, self._size)
        if slice_stop > len(self):
            logger.warning(
                "Specified slice has a start %d equal to its stop %d "
                "(based on size %d). Therefore slice will be empty",
                slice_start, slice_stop, self._size)
        if slice_stop < 0:
            logger.warning(
                "Specified slice stop was %d while size is only %d. "
                "Therefore slice will be empty",
                int(slice_stop) - self._size, self._size)
            return (self._size, self._size)
        elif slice_start > slice_stop:
            logger.warning(
                "Specified slice has a start %d greater than its stop %d "
                "(based on size %d). Therefore slice will be empty",
                slice_start, slice_stop, self._size)
            return (self._size, self._size)
        elif slice_start == slice_stop:
            logger.warning(
                "Specified slice has a start %d equal to its stop %d "
                "(based on size %d). Therefore slice will be empty",
                slice_start, slice_stop, self._size)
        elif slice_stop > len(self):
            logger.warning(
                "Specified slice stop was %d while size is only %d. "
                "Therefore slice will be truncated",
                slice_stop, self._size)
            slice_stop = self._size
        return slice_start, slice_stop

    def _check_mask_size(self, selector: Sized) -> None:
        if len(selector) < self._size:
            logger.warning(
                "The boolean mask is too short. The expected length was %d "
                "but the length was only %d. All the missing entries will be "
                "treated as False!", self._size, len(selector))
        elif len(selector) > self._size:
            logger.warning(
                "The boolean mask is too long. The expected length was %d "
                "but the length was only %d. All the missing entries will be "
                "ignored!", self._size, len(selector))

    def selector_to_ids(self, selector: Selector, warn=False) -> Sequence[int]:
        """
        Gets the list of IDs covered by this selector.
        The types of selector currently supported are:

        `None`:
            Returns all IDs.

        :py:class:`slice`: Standard python slice.
            Negative values and values larger than size are handled using
            slice's `indices` method.
            This could result in am empty list.

        :py:class:`int`: Handles negative values as normal.
            Checks if ID is within expected range.

        iterator(:py:class:`bool`): Used as a mask.
            If the length of the mask is longer or shorted than number of IDs
            the result is the shorter of the two,
            with the remainder of the longer ignored.

        iterator(:py:class:`int`) but not bool:
            Every value checked that it is with the range 0 to size.
            Negative values are *not* allowed.
            Original order and duplication is respected so result may be
            unordered and contain duplicates.

        :param selector: Some object that identifies a range of IDs.
        :param bool warn:
            If True, this method will warn about problems with the selector.
        :return: a (possibly sorted) list of IDs
        """
        if _is_iterable_selector(selector):
            # bool is subclass of int so if any are bool all must be
            if any(isinstance(item, (bool, numpy.bool_)) for item in selector):
                if all(isinstance(item, (bool, numpy.bool_))
                       for item in selector):
                    if warn:
                        self._check_mask_size(selector)
                    return list(itertools.compress(
                        range(self._size), selector))
                raise TypeError(
                    "An iterable type must be all ints or all bools")
            elif all(isinstance(item, (int, numpy.integer, SupportsInt))
                     for item in selector):
                # list converts any specific numpy types
                # blows up if someone gives an iterable of ndarray;
                # serves them right for being too smart-ass
                ids = list(map(int, selector))
                for _id in ids:
                    if _id < 0:
                        raise TypeError(
                            f"Selector includes the ID {_id} which is "
                            "less than zero")
                    if _id >= self._size:
                        raise TypeError(
                            f"Selector includes the ID {_id} which not "
                            f"less than the size {self._size}")
                return ids
            else:
                raise TypeError(
                    "An iterable type must be all ints or all bools")

        # OK lets try for None, int and slice after all
        if selector is None:
            if warn:
                logger.warning("None selector taken as all IDs")
            return range(self._size)

        if isinstance(selector, slice):
            (slice_start, slice_stop, step) = selector.indices(self._size)
            return range(slice_start, slice_stop, step)

        if isinstance(selector, (int, numpy.integer)):
            selector = int(selector)  # make sure it is a simple int
            if selector < 0:
                selector = self._size + selector
            if selector < 0 or selector >= self._size:
                raise TypeError(
                    f"Selector {selector-self._size} is unsupported "
                    f"for size {self._size}")
            return [selector]

        # Last throw of the dice...
        if isinstance(selector, SupportsInt):
            selector = int(selector)
            if selector < 0:
                selector = self._size + selector
            if selector < 0 or selector >= self._size:
                raise TypeError(
                    f"Selector {selector-self._size} is unsupported "
                    f"for size {self._size}")
            return [selector]

        raise TypeError(f"Unexpected selector type {type(selector)}")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class SafeEval(object):
    """
    This provides expression evaluation capabilities while allowing the
    set of symbols exposed to the expression to be strictly controlled.

    Sample of use::

        >>> import math
        >>> def evil_func(x):
        ...    print("Hello World!")
        ...    return x/0.0
        ...
        >>> eval_safely = SafeEval(math)
        >>> eval_safely.eval("math.sqrt(x)", x=1.23)
        1.1090536506409416
        >>> eval_safely.eval("evil_func(1.23)")
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
          File ".../safe_eval.py", line 62, in eval
            return eval(expression, self._environment, kwargs)
          File "<string>", line 1, in <module>
        NameError: name 'evil_func' is not defined

    .. warning::
        This is not guaranteed to be safe under all circumstances. It is not
        designed to be a fully secured interface; it just *discourages* abuse.
    """
    __slots__ = ["_environment"]

    def __init__(self, *args, **kwargs):
        """
        :param args:
            The symbols to use to populate the global reference table.

            .. note::
                All of these symbols must support the `__name__` property, but
                that includes any function, method of an object, or module. If
                you want to make an object available by anything other than its
                inherent name, define it in the
                :py:meth:`eval` call.
        :param kwargs:
            Define the symbols with explicit names. Needed because some
            symbols (e.g., constants in numpy) do not have names that we can
            otherwise look up easily.
        """
        env = {}
        for item in args:
            env[item.__name__] = item
        env.update(kwargs)
        self._environment = env

    def eval(self, expression, **kwargs):
        """
        Evaluate an expression and return the result.

        :param expression: The expression to evaluate
        :type expression: str
        :param kwargs:
            The extra symbol bindings to use for this evaluation.
            This is useful for passing in particular parameters to an
            individual evaluation run.
        :return: The expression result, the type of which will depend on the
            expression itself and the operations exposed to it.
        """
        # pylint: disable=eval-used
        return eval(expression, self._environment, kwargs)



# Copyright (c) 2018 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
import os
import re
from spinn_utilities.exceptions import UnexpectedCException
from .log_sqllite_database import LogSqlLiteDatabase

TOKEN = chr(30)  # Record Separator

COMMA_SPLIITER = re.compile(r'(?!\B"[^"]*),(?![^"]*"\B)')
STRING_REGEXP = re.compile(r'"([^"]|\\"|(""))*"')
FORMAT_EXP = re.compile(r"(%+\d*(?:\.\d+)?[cdfiksuxRFK])")
LOG_END_REGEX = re.compile(r'\)(\s)*;')
END_COMMENT_REGEX = re.compile(r"/*/")
LOG_START_REGEX = re.compile(
    r"log_((info)|(error)|(debug)|(warning))(\s)*\(")
DOUBLE_HEX = ", double_to_upper({0}), double_to_lower({0})"
LEVELS = {"log_info(": 20,
          "log_error(": 40,
          "log_debug(": 10,
          "log_warning(": 30}

MINIS = {"log_info(": "log_mini_info(",
         "log_error(": "log_mini_error(",
         "log_debug(": "log_mini_debug(",
         "log_warning(": "log_mini_warning("}


class State(enum.Enum):
    """
    Status values.
    """
    NORMAL_CODE = 0
    COMMENT = 1
    IN_LOG = 2
    IN_LOG_CLOSE_BRACKET = 3


class FileConverter(object):
    """
    Converts a file. See :py:meth:`convert`.
    """

    __slots__ = [
        "_log_database",
        "_log_file_id",
        "_log",
        "_log_full",
        "_log_lines",
        "_log_start",
        "_previous_status",
        "_src",
        "_status",
        "_too_many_lines"
    ]

    def __call__(self, src, dest, log_file_id, log_database):
        """
        Creates the file_convertor to convert one file.

        :param str src: Absolute path to source file
        :param str dest: Absolute path to destination file
        :param int log_file_id:
            Id in the database for this file
        :param LogSqlLiteDatabase log_database:
            The database which handles the mapping of id to log messages.
        """
        #: Absolute path to source file
        #:
        #: :type: str
        self._src = src
        #: Database which handles the mapping of id to log messages
        #:
        #: :type: .log_sqllite_database.LogSqlLiteDatabase
        self._log_database = log_database
        #: Id in the database for this file
        #:
        #: :type: int
        self._log_file_id = log_file_id
        #: Current status of state machine
        #:
        #: :type: State
        self._status = None
        #: Number of extra lines written to modified not yet recovered
        #: Extra lines are caused by the header and possibly log comment
        #: Extra lines are recovered by omitting blank lines
        #:
        #: :type: int
        self._too_many_lines = None
        #: Variables created each time a log method found
        #: original c log method found
        #:
        #: :type: str
        self._log = None
        #: Log methods found so far
        #:
        #: :type: str
        self._log_full = None
        #: Number of c lines the log method takes
        #:
        #: :type: int
        self._log_lines = None
        #: Any other stuff found before the log method but on same line
        #:
        #: :type: str
        self._log_start = None
        # variable created when a comment found
        #: The previous state
        #:
        #: :type: State
        self._previous_status = None

        with open(src, encoding="utf-8") as src_f:
            with open(dest, 'w', encoding="utf-8") as dest_f:
                dest_f.write(
                    f"// DO NOT EDIT! THIS FILE WAS GENERATED FROM "
                    f"{os.path.relpath(src, dest)}\n\n")
                self._too_many_lines = 2
                self._status = State.NORMAL_CODE
                for line_num, text in enumerate(src_f):
                    if self._too_many_lines > 0:
                        # Try to recover the lines added by do not edit
                        check = text.strip()
                        if len(check) == 0 or check == "*":
                            self._too_many_lines -= 1
                            continue
                    previous_status = self._status
                    if not self._process_line(dest_f, line_num, text):
                        self._status = previous_status
                        self._process_chars(dest_f, line_num, text)
        self._check_end_status()

    def _check_end_status(self):
        if self._status == State.NORMAL_CODE:
            return
        if self._status == State.IN_LOG:
            raise UnexpectedCException(
                f"Unclosed {self._log}{self._log_full} in {self._src}")
        if self._status == State.IN_LOG_CLOSE_BRACKET:
            raise UnexpectedCException(
                f"Semicolumn missing: "
                f"{self._log}{self._log_full} in {self._src}")
        if self._status == State.COMMENT:
            raise UnexpectedCException(
                f"Unclosed block comment in {self._src}")
        raise NotImplementedError(f"Unexpected status {self._status}")

    def _process_line(self, dest_f, line_num, text):
        """
        Process a single line.

        :param dest_f: Open file like Object to write modified source to
        :param int line_num: Line number in the source c file
        :param str text: Text of that line including whitespace
        :return: True if and only if the whole line was processed
        :rtype: bool
        """
        if self._status == State.COMMENT:
            return self._process_line_in_comment(dest_f, text)
        if "/*" in text:
            return self._process_line_comment_start(dest_f, line_num, text)

        if self._status == State.IN_LOG:
            return self._process_line_in_log(dest_f, line_num, text)

        if self._status == State.IN_LOG_CLOSE_BRACKET:
            return self._process_line_in_log_close_bracket(
                dest_f, line_num, text)

        assert self._status == State.NORMAL_CODE
        return self._process_line_normal_code(dest_f, line_num, text)

    def _process_line_in_comment(self, dest_f, text):
        """
        Process a single line when in a multi-line comment: ``/* .. */``

        :param dest_f: Open file like Object to write modified source to
        :param str text: Text of that line including whitespace
        :return: True if and only if the whole line was processed
        :rtype: bool
        """
        if "*/" in text:
            stripped = text.strip()
            match = END_COMMENT_REGEX.search(stripped)
            if match.end(0) == len(stripped):
                # OK Comment until end of line
                dest_f.write(text)
                self._status = State.NORMAL_CODE
                return True
            return False  # Stuff after comment so check by character
        # Whole line in comment without end
        dest_f.write(text)
        return True

    def _process_line_comment_start(self, dest_f, line_num, text):
        """
        Processes a line known assumed to contain a ``/*`` but not know where.

        There is also the assumption that the start status is not ``COMMENT``.

        :param dest_f: Open file like Object to write modified source to
        :param int line_num: Line number in the source c file
        :param str text: Text of that line including whitespace
        :return: True if and only if the whole line was processed
        :rtype: bool
        """
        stripped = text.strip()
        if stripped.startswith("/*"):
            self._previous_status = self._status
            self._status = State.COMMENT
            # Comment start so now check for comment end
            return self._process_line(dest_f, line_num, text)
        # Stuff before comment so check by char
        return False  # More than one possible end so check by char

    def _process_line_in_log(self, dest_f, line_num, text):
        """
        Process a line when the status is a log call has been started.

        :param dest_f: Open file like Object to write modified source to
        :param int line_num: Line number in the source c file
        :param str text: Text of that line including whitespace
        :return: True if and only if the whole line was processed
        :rtype: bool
        """
        stripped = text.strip()
        if stripped.startswith("//"):
            # Just a comment line so write and move on
            dest_f.write(text)
            return True

        match = LOG_END_REGEX.search(stripped)
        if not match:
            if stripped[-1:] == ")":
                # possible start of end
                self._status = State.IN_LOG_CLOSE_BRACKET
            self._log_full += stripped
            self._log_lines += 1
            return True
        if match.end(0) < len(stripped):
            # Stuff after the log_end so check by char
            return False

        self._log_lines += 1
        self._log_full += stripped
        self._write_log_method(dest_f, line_num)
        self._status = State.NORMAL_CODE
        return True

    def _process_line_in_log_close_bracket(self, dest_f, line_num, text):
        """
        Process where the last log line has the ``)`` but not the ``;``

        :param dest_f: Open file like Object to write modified source to
        :param int line_num: Line number in the source c file
        :param str text: Text of that line including whitespace
        :return: True if and only if the whole line was processed
        :rtype: bool
        """
        stripped = text.strip()
        if len(stripped) == 0:
            self._log_lines += 1
            return True
        if stripped[0] == ";":
            if stripped == ";":
                self._log_full += (";")
                self._log_lines += 1
                self._write_log_method(dest_f, line_num)
                self._status = State.NORMAL_CODE
                return True
            else:
                return False

        elif stripped.startswith("//"):
            # Just a comment line so write and move on
            dest_f.write(text)
            return True

        else:
            # so not a closing bracket so set status back
            self._status = State.IN_LOG
            return self._process_line_in_log(dest_f, line_num, text)

    def _process_line_normal_code(self, dest_f, line_num, text):
        """
        Process a line where the status is normal code.

        :param dest_f: Open file like Object to write modified source to
        :param int line_num: Line number in the source c file
        :param str text: Text of that line including whitespace
        :return: True if and only if the whole line was processed
        :rtype: bool
        """
        stripped = text.strip()
        match = LOG_START_REGEX.search(stripped)
        if not match:
            # No log start found after all
            dest_f.write(text)
            return True

        if match.start() > 0:
            if stripped.startswith("//"):
                # Just a comment line so write and move on
                dest_f.write(text)
                return True
            # Stuff before the log_start so check by character
            return False

        if LOG_START_REGEX.search(stripped, 1):
            # Second start found so check by character
            return False

        # remove white spaces and save log command
        self._log_start = text.index(match.group(0))
        self._log = "".join(match.group(0).split())
        start_len = self._log_start + len(self._log)

        self._status = State.IN_LOG
        self._log_full = ""  # text saved in process_line_in_log
        self._log_lines = 0
        # Now check for the end of log command
        return self._process_line_in_log(dest_f, line_num, text[start_len:])

    def quote_part(self, text):
        """
        Net count of double quotes in line.

        :param str text:
        :rtype: int
        """
        return (text.count('"') - text.count('\\"')) % 2 > 0

    def bracket_count(self, text):
        """
        Net count of open brackets in line.

        :param str text:
        :rtype: int
        """
        return (text.count('(') - text.count(')'))

    def split_by_comma_plus(self, main, line_num):
        """
        Split line by comma and partially parse.

        :param str main:
        :param int line_num:
        :rtype: list(str)
        :raises UnexpectedCException:
        """
        try:
            parts = main.split(",")
            for i, part in enumerate(parts):
                check = part.strip()
                if check[0] == '"':
                    # Dealing with a String
                    if check[-1] == '"':
                        if check[-2] != '\\':
                            # Part is a full sting fine
                            continue
                    # Save start of String and get next part
                    new_part = parts.pop(i)
                    next_part = parts.pop(i)
                    new_check = next_part.strip()
                    while new_check[-1] != '"' or new_check[-2] == '\\':
                        # Still not end of String so add and get next
                        new_part += "," + next_part
                        next_part = parts.pop(i)
                        new_check = next_part.strip()
                    # Add the end and put back new in the list
                    new_part += "," + next_part
                    parts.insert(i, new_part)
                else:
                    # Not a String so look for function
                    count = self.bracket_count(part)
                    if count > 0:
                        # More opening and closing brackets so in function
                        new_part = parts.pop(i)
                        # Keep combining parts until you find the last closing
                        while count > 0:
                            next_part = parts.pop(i)
                            count += self.bracket_count(next_part)
                            new_part += "," + next_part
                        # Put the new part back into the list
                        parts.insert(i, new_part)
            if parts[0][0] == '"' and parts[0][-1] == '"':
                parts[0] = parts[0][1:-1]
            return parts

        except Exception as e:
            raise UnexpectedCException(f"Unexpected line {self._log_full} "
                                       f"at {line_num} in {self._src}") from e

    def _short_log(self, line_num):
        """
        Shortens the log string message and adds the ID.

        :param int line_num: Current line number
        :return: shorten form
        :rtype: str
        """
        try:
            match = LOG_END_REGEX.search(self._log_full)
            main = self._log_full[:-len(match.group(0))]
        except Exception as e:
            raise UnexpectedCException(
                f"Unexpected line {self._log_full} at "
                f"{line_num} in {self._src}") from e
        parts = self.split_by_comma_plus(main, line_num)
        original = parts[0]

        message_id = self._log_database.set_log_info(
            LEVELS[self._log], line_num + 1, original, self._log_file_id)
        count = original.count("%") - original.count("%%") * 2

        if count == 0:
            return f'"%u", {message_id});'

        front = '"%u'
        back = ""
        matches = [x for x in FORMAT_EXP.findall(original)
                   if not x.startswith("%%")]
        if len(matches) != count:
            raise UnexpectedCException(
                f"Unexpected formatString in {original}")
        if len(parts) < count + 1:
            raise UnexpectedCException(
                f"Too few parameters in line {self._log_full} "
                f"at {line_num} in {self._src}")
        if len(parts) > count + 1:
            raise UnexpectedCException(
                f"Too many parameters in line {self._log_full} "
                f"at {line_num} in {self._src}")
        for i, match in enumerate(matches):
            front += TOKEN
            if match.endswith("f"):
                front += "%x"
                back += f", float_to_int({parts[i + 1]})"
            elif match.endswith("F"):
                front += "%x" + TOKEN + "%x"
                back += DOUBLE_HEX.format(parts[i + 1])
            else:
                back += f", {parts[i + 1]}"
                front += match

        front += f'", {message_id}'
        back += ");"
        return front + back

    def _write_log_method(self, dest_f, line_num, tail=""):
        """
        Writes the log message and the dict value.

        Writes the log call to the destination
        - New log method used
        - Shortened log message (with just an id) used
        - Parameters kept as is
        - Old log message with full text added as comment

        :param dest_f: Open file like Object to write modified source to
        :param int line_num: Line number in the source C file
        :param str text: Text of that line including whitespace
        """
        self._log_full = self._log_full.replace('""', '')
        short_log = self._short_log(line_num)

        dest_f.write(" " * self._log_start)
        dest_f.write(MINIS[self._log])
        dest_f.write(short_log)
        if self._log_lines == 0:
            # Writing an extra newline here so need to recover that ASAP
            self._too_many_lines += 1
        end = tail + "\n"
        if self._log_lines <= 1:
            dest_f.write("  /* ")
            dest_f.write(self._log)
            dest_f.write(self._log_full)
            dest_f.write("*/")
            dest_f.write(end)
        else:
            dest_f.write(tail)
            dest_f.write(end)
            dest_f.write(" " * self._log_start)
            dest_f.write("/* ")
            dest_f.write(self._log)
            dest_f.write(self._log_full)
            dest_f.write("*/")
            dest_f.write(end * (self._log_lines - 1))

    def _process_chars(self, dest_f, line_num, text):
        """
        Deals with complex lines that can not be handled in one go.

        :param dest_f: Open file like Object to write modified source to
        :param int line_num: Line number in the source c file
        :param str text: Text of that line including whitespace
        :raises UnexpectedCException:
        """
        position = 0
        write_flag = 0
        while text[position] != "\n":
            if self._status == State.COMMENT:
                if text[position] == "*" and text[position+1] == "/":
                    dest_f.write(text[write_flag:position + 2])
                    position = position + 2
                    write_flag = position
                    self._status = self._previous_status
                else:
                    position = position + 1
            elif text[position] == "/":
                if text[position+1] == "*":
                    if self._status == State.IN_LOG:
                        self._log_full += text[write_flag:position].strip()
                        if self._log_full[-1] == ")":
                            self._status = State.IN_LOG_CLOSE_BRACKET
                        # NO change to self._log_lines as newline not removed
                    else:
                        dest_f.write(text[write_flag:position])
                    write_flag = position
                    position = position + 2  # leave the /* as not written
                    self._previous_status = self._status
                    self._status = State.COMMENT
                elif text[position+1] == "/":
                    if self._status == State.IN_LOG:
                        self._log_full += text[write_flag:position].strip()
                        # NO change to self._log_lines as newline not removed
                        dest_f.write(text[position:])
                    else:
                        dest_f.write(text[write_flag:])
                    return  # Finished line
                else:
                    position += 1

            elif text[position] == '"':
                str_pos = position + 1
                while text[str_pos] != '"':
                    if text[str_pos] == "\n":
                        raise UnexpectedCException(
                            f"Unclosed string literal in {self._src} "
                            f"at line: {line_num}")
                    elif text[str_pos] == "\\":
                        if text[str_pos+1] == "\n":
                            raise UnexpectedCException(
                                f"Unclosed string literal in {self._src} "
                                f"at line: {line_num}")
                        else:
                            str_pos += 2  # ignore next char which may be a "
                    else:
                        str_pos += 1
                position = str_pos + 1
                continue

            elif self._status == State.IN_LOG:
                if text[position] == ")":
                    match = LOG_END_REGEX.match(text[position:])
                    if match:
                        # include the end
                        position = position + len(match.group(0))
                        self._log_full += text[write_flag:position].strip()
                        self._status = State.NORMAL_CODE
                        if text[position:].strip():  # Stuff left
                            write_flag = position
                            # self._log_lines not changed as no newline
                            # check for a \ after the log
                            if text[position:].strip() == "\\":
                                self._write_log_method(dest_f, line_num, "\\")
                                return
                            else:
                                self._write_log_method(dest_f, line_num)
                        else:
                            self._log_lines += 1
                            self._write_log_method(dest_f, line_num)
                            return  # Finished line
                    else:
                        # not the require ); so continue
                        position += 1
                else:
                    position += 1

            elif self._status == State.IN_LOG_CLOSE_BRACKET:
                stripped = text.strip()
                if stripped[0] == ";":
                    self._log_full += (";")
                    self._write_log_method(dest_f, line_num)
                    position = text.index(";") + 1
                    write_flag = position
                    self._status = State.NORMAL_CODE
                else:
                    # Save the ) as not part of the end
                    self._status = State.IN_LOG

            elif text[position] == "l":
                match = LOG_START_REGEX.match(text[position:])
                if match:
                    self._log_start = text.index(match.group(0))
                    self._log = "".join(match.group(0).split())
                    self._status = State.IN_LOG
                    self._log_full = ""  # text saved after while
                    self._log_lines = 0
                    dest_f.write(text[write_flag:position])
                    # written up to not including log_start
                    # skip to end of log instruction
                    position = position + len(match.group(0))
                    write_flag = position
                else:
                    # Not a log start after all so treat as normal test
                    position += 1

            else:
                position += 1

        # after while text[position] != "\n"
        if self._status == State.IN_LOG:
            self._log_full += text[write_flag:].strip()
            self._log_lines += 1
        else:
            dest_f.write(text[write_flag:])

    @staticmethod
    def convert(src_dir, dest_dir, file_name):
        """
        Static method to create Object and do the conversion.

        :param str src_dir: Source directory
        :param str dest_dir: Destination directory
        :param str file_name:
            The name of the file to convert within the source directory; it
            will be made with the same name in the destination directory.
        """
        source = os.path.join(src_dir, file_name)
        if not os.path.exists(source):
            raise UnexpectedCException(f"Unable to locate source {source}")
        if not os.path.exists(dest_dir):
            os.makedirs(dest_dir)
        destination = os.path.join(dest_dir, file_name)
        with LogSqlLiteDatabase() as log_database:
            directory_id = log_database.get_directory_id(src_dir, dest_dir)
            file_id = log_database.get_file_id(directory_id, file_name)
            FileConverter()(source, destination, file_id, log_database)



# Copyright (c) 2018 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional


def as_string(byte_string: bytes, start: Optional[int] = None,
              end: Optional[int] = None) -> str:
    """
    Returns the length and the hex values.

    The length is always the full length irrespective of the start and end.

    :param bytes byte_string: data as a byte string
    :param int start:
        The inclusive start of the slice to convert to hexadecimal.
        May be `None`
    :param int end:
        The exclusive end of the slice to convert to hexadecimal. May be `None`
    :return:
        The length of the byte string and the comma separated hex values, as a
        descriptive string
    :rtype: str
    """
    return "(" + str(len(byte_string)) + ")" + as_hex(byte_string, start, end)


def as_hex(byte_string: bytes, start: Optional[int] = None,
           end: Optional[int] = None) -> str:
    """
    Returns the byte string as string showing the hex values

    :param bytes byte_string: data as a byte string
    :param int start: the inclusive start of the slice to return. May be `None`
    :param int end: the exclusive end of the slice to return. May be `None`
    :return: Comma-separated hex values
    :rtype: str
    """
    # pylint: disable=consider-using-f-string
    return ','.join('%02x' % i for i in iter(byte_string[start:end]))



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import logging
import os
from typing import Any, Callable, Collection, Dict, List, Optional, Set, Union
import spinn_utilities.conf_loader as conf_loader
from spinn_utilities.configs import CamelCaseConfigParser
from spinn_utilities.exceptions import ConfigException
from spinn_utilities.log import FormatAdapter

# pylint: disable=global-statement
logger = FormatAdapter(logging.getLogger(__file__))

__config: Optional[CamelCaseConfigParser] = None
__default_config_files: List[str] = []
__config_file: Optional[str] = None
__unittest_mode: bool = False


def add_default_cfg(default: str):
    """
    Adds an extra default configuration file to be read after earlier ones.

    :param str default: Absolute path to the configuration file
    """
    if default not in __default_config_files:
        __default_config_files.append(default)


def clear_cfg_files(unittest_mode: bool):
    """
    Clears any previous set configurations and configuration files.

    After this method :py:func:`add_default_cfg` and :py:func:`set_cfg_files`
    need to be called.

    :param bool unittest_mode: Flag to put the holder into unit testing mode
    """
    global __config, __config_file, __unittest_mode
    __config = None
    __default_config_files.clear()
    __config_file = None
    __unittest_mode = unittest_mode


def set_cfg_files(config_file: str, default: str):
    """
    Adds the configuration files to be loaded.

    :param str config_file:
        The base name of the configuration file(s).
        Should not include any path components.
    :param str default:
        Full path to the extra file to get default configurations from.
    """
    global __config_file
    __config_file = config_file
    add_default_cfg(default)


def _pre_load_config() -> CamelCaseConfigParser:
    """
    Loads configurations due to early access to a configuration value.

    :raises ConfigException: Raise if called before setup
    """
    # If you get this error during a unit test, unittest_step was not called
    if not __unittest_mode:
        raise ConfigException(
            "Accessing config values before setup is not supported")
    return load_config()


def load_config() -> CamelCaseConfigParser:
    """
    Reads in all the configuration files, resetting all values.

    :raises ConfigException: If called before setting defaults
    """
    global __config
    if not __default_config_files:
        raise ConfigException("No default configs set")
    if __config_file:
        __config = conf_loader.load_config(
            filename=__config_file, defaults=__default_config_files)
    else:
        __config = CamelCaseConfigParser()
        for default in __default_config_files:
            __config.read(default)
    return __config


def is_config_none(section, option) -> bool:
    """
    Check if the value of a configuration option would be considered None

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :return: True if and only if the value would be considered None
    :rtype: bool
    """
    value = get_config_str_or_none(section, option)
    return value is None


def get_config_str(section, option) -> str:
    """
    Get the string value of a configuration option.

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :return: The option value
    :rtype: str
    :raises ConfigException: if the Value would be None
    """
    value = get_config_str_or_none(section, option)
    if value is None:
        raise ConfigException(f"Unexpected None for {section=} {option=}")
    return value


def get_config_str_or_none(section, option) -> Optional[str]:
    """
    Get the string value of a configuration option.

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :return: The option value
    :rtype: str or None
    :raises ConfigException: if the Value would be None
    """
    if __config is None:
        return _pre_load_config().get_str(section, option)
    else:
        return __config.get_str(section, option)


def get_config_str_list(
        section: str, option: str, token: str = ",") -> List[str]:
    """
    Get the string value of a configuration option split into a list.

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :param token: The token to split the string into a list
    :return: The list (possibly empty) of the option values
    :rtype: list(str)
    """
    if __config is None:
        return _pre_load_config().get_str_list(section, option, token)
    else:
        return __config.get_str_list(section, option, token)


def get_config_int(section: str, option: str) -> int:
    """
    Get the integer value of a configuration option.

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :return: The option value
    :rtype: int
    :raises ConfigException: if the Value would be None
    """
    value = get_config_int_or_none(section, option)
    if value is None:
        raise ConfigException(f"Unexpected None for {section=} {option=}")
    return value


def get_config_int_or_none(section, option) -> Optional[int]:
    """
    Get the integer value of a configuration option.

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :return: The option value
    :rtype: int or None
    :raises ConfigException: if the Value would be None
    """
    if __config is None:
        return _pre_load_config().get_int(section, option)
    else:
        return __config.get_int(section, option)


def get_config_float(section: str, option: str) -> float:
    """
    Get the float value of a configuration option.

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :return: The option value.
    :rtype: float
    :raises ConfigException: if the Value would be None
    """
    value = get_config_float_or_none(section, option)
    if value is None:
        raise ConfigException(f"Unexpected None for {section=} {option=}")
    return value


def get_config_float_or_none(section, option) -> Optional[float]:
    """
    Get the float value of a configuration option.

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :return: The option value.
    :rtype: float or None
    """
    if __config is None:
        return _pre_load_config().get_float(section, option)
    else:
        return __config.get_float(section, option)


def get_config_bool(section: str, option: str) -> bool:
    """
    Get the Boolean value of a configuration option.

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :return: The option value.
    :rtype: bool
    :raises ConfigException: if the Value would be None
    """
    value = get_config_bool_or_none(section, option)
    if value is None:
        raise ConfigException(f"Unexpected None for {section=} {option=}")
    return value


def get_config_bool_or_none(section, option) -> Optional[bool]:
    """
    Get the Boolean value of a configuration option.

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :return: The option value.
    :rtype: bool
    :raises ConfigException: if the Value would be None
    """
    if __config is None:
        return _pre_load_config().get_bool(section, option)
    else:
        return __config.get_bool(section, option)


def set_config(section: str, option: str, value: Optional[str]):
    """
    Sets the value of a configuration option.

    This method should only be called by the simulator or by unit tests.

    :param str section: What section to set the option in.
    :param str option: What option to set.
    :param object value: Value to set option to
    :raises ConfigException: If called unexpectedly
    """
    if __config is None:
        _pre_load_config().set(section, option, value)
    else:
        __config.set(section, option, value)


def has_config_option(section: str, option: str) -> bool:
    """
    Check if the section has this configuration option.

    :param str section: What section to check
    :param str option: What option to check.
    :rtype: bool
    :return: True if and only if the option is defined. It may be `None`
    """
    if __config is None:
        raise ConfigException("configuration not loaded")
    else:
        return __config.has_option(section, option)


def config_options(section: str) -> List[str]:
    """
    Return a list of option names for the given section name.

    :param str section: What section to list options for.
    """
    if __config is None:
        raise ConfigException("configuration not loaded")
    return __config.options(section)


def _check_lines(py_path: str, line: str, lines: List[str], index: int,
                 method: Callable[[str, str], Any],
                 used_cfgs: Dict[str, Set[str]], start):
    """
    Support for `_check_python_file`. Gets section and option name.

    :param str line: Line with get_config call
    :param list(str) lines: All lines in the file
    :param int index: index of line with `get_config` call
    :param method: Method to call to check cfg
    :param dict(str), set(str) used_cfgs:
        Dict of used cfg options to be added to
    :raises ConfigException: If an unexpected or uncovered `get_config` found
    """
    while ")" not in line:
        index += 1
        line += lines[index]
    parts = line[line.find("(", line.find(start)) + 1:
                 line.find(")")].split(",")
    section = parts[0].strip().replace("'", "").replace('"', '')
    for i in range(1, len(parts)):
        try:
            option = parts[i].strip()
        except IndexError as original:
            raise ConfigException(
                f"failed in line:{index} of file: {py_path} with {line}") \
                from original
        if option[0] == "'":
            option = option.replace("'", "")
        elif option[0] == '"':
            option = option.replace('"', '')
        else:
            print(line)
            return
        try:
            method(section, option)
        except Exception as original:
            raise ConfigException(
                f"failed in line:{index} of file: {py_path} with "
                f"section:{section} option:{option}") from original
        used_cfgs[section].add(option)


def _check_python_file(py_path: str, used_cfgs: Dict[str, Set[str]]):
    """
    A testing function to check that all the `get_config` calls work.

    :param str py_path: path to file to be checked
    :param used_cfgs: dict of cfg options found
    :raises ConfigException: If an unexpected or uncovered `get_config` found
    """
    with open(py_path, 'r', encoding="utf-8") as py_file:
        lines = list(py_file)
        for index, line in enumerate(lines):
            if ("skip_if_cfg" in line):
                _check_lines(py_path, line, lines, index,
                             get_config_bool_or_none, used_cfgs, "skip_if_cfg")
            if ("configuration.get" in line):
                _check_lines(py_path, line, lines, index,
                             get_config_bool_or_none, used_cfgs,
                             "configuration.get")
            if "get_config" not in line:
                continue
            if (("get_config_bool(" in line) or
                    ("get_config_bool_or_none(" in line)):
                _check_lines(py_path, line, lines, index,
                             get_config_bool_or_none, used_cfgs, "get_config")
            if (("get_config_float(" in line) or
                    ("get_config_float_or_none(" in line)):
                _check_lines(py_path, line, lines, index,
                             get_config_float_or_none, used_cfgs, "get_config")
            if (("get_config_int(" in line) or
                    ("get_config_int_or_none(" in line)):
                _check_lines(py_path, line, lines, index,
                             get_config_int_or_none, used_cfgs, "get_config")
            if (("get_config_str(" in line) or
                    ("get_config_str_or_none(" in line)):
                _check_lines(py_path, line, lines, index,
                             get_config_str_or_none, used_cfgs, "get_config")
            if "get_config_str_list(" in line:
                _check_lines(py_path, line, lines, index,
                             get_config_str_list, used_cfgs, "get_config")


def _find_double_defaults(repeaters: Optional[Collection[str]] = ()):
    """
    Testing function to identify any configuration options in multiple default
    files.

    :param repeaters: List of options that are expected to be repeated.
    :type repeaters: list(str)
    :raises ConfigException:
        If two defaults configuration files set the same value
    """
    config1 = CamelCaseConfigParser()
    for default in __default_config_files[:-1]:
        config1.read(default)
    config2 = CamelCaseConfigParser()
    config2.read(__default_config_files[-1])
    if repeaters is None:
        repeaters = []
    else:
        repeaters = frozenset(map(config2.optionxform, repeaters))
    for section in config2.sections():
        for option in config2.options(section):
            if config1.has_option(section, option):
                if option not in repeaters:
                    raise ConfigException(
                        f"cfg:{__default_config_files[-1]} "
                        f"repeats [{section}]{option}")


def _check_cfg_file(config1: CamelCaseConfigParser, cfg_path: str):
    """
    Support method for :py:func:`check_cfgs`.

    :param CamelCaseConfigParser config1:
    :param str cfg_path:
    :raises ConfigException: If an unexpected option is found
    """
    config2 = CamelCaseConfigParser()
    config2.read(cfg_path)
    for section in config2.sections():
        if not config1.has_section(section):
            raise ConfigException(
                f"cfg:{cfg_path} has unexpected section [{section}]")
        for option in config2.options(section):
            if not config1.has_option(section, option):
                raise ConfigException(
                    f"cfg:{cfg_path} "
                    f"has unexpected options [{section}]{option}")


def _check_cfgs(path: str):
    """
    A testing function check local configuration files against the defaults.

    It only checks that the option exists in a default.
    It does not check if the option is used or if the value is the expected
    type.

    :param str path: Absolute path to the parent directory to search
    :raises ConfigException: If an unexpected option is found
    """
    config1 = CamelCaseConfigParser()
    for default in __default_config_files:
        config1.read(default)
    directory = os.path.dirname(path)
    for root, _, files in os.walk(directory):
        for file_name in files:
            if file_name.endswith(".cfg"):
                cfg_path = os.path.join(root, file_name)
                if cfg_path in __default_config_files:
                    continue
                print(cfg_path)
                _check_cfg_file(config1, cfg_path)


def run_config_checks(directories: Union[str, Collection[str]], *,
                      exceptions: Union[str, Collection[str]] = (),
                      repeaters: Optional[Collection[str]] = (),
                      check_all_used: bool = True):
    """
    Master test.

    Checks that all cfg options read have a default value in one of the
    default files.

    Checks that all default options declared in the current repository
    are used in that repository.

    :param module:
    :param exceptions:
    :param repeaters:
    :param bool check_all_used: Toggle for the used test.
    :raises ConfigException: If an incorrect directory passed in
    """
    if isinstance(directories, str):
        directories = [directories]

    if exceptions is None:
        exceptions = []
    elif isinstance(exceptions, str):
        exceptions = [exceptions]

    _find_double_defaults(repeaters)

    config1 = CamelCaseConfigParser()
    config1.read(__default_config_files)

    used_cfgs: Dict[str, Set[str]] = defaultdict(set)
    for directory in directories:
        if not os.path.isdir(directory):
            raise ConfigException(f"Unable find {directory}")
        for root, _, files in os.walk(directory):
            for file_name in files:
                if file_name in exceptions:
                    pass
                elif file_name.endswith(".cfg"):
                    cfg_path = os.path.join(root, file_name)
                    if cfg_path in __default_config_files:
                        continue
                    print(cfg_path)
                    _check_cfg_file(config1, cfg_path)
                elif file_name.endswith(".py"):
                    py_path = os.path.join(root, file_name)
                    _check_python_file(py_path, used_cfgs)

    if not check_all_used:
        return

    config2 = CamelCaseConfigParser()
    config2.read(__default_config_files[-1])
    for section in config2:
        if section not in used_cfgs:
            if section == config1.default_section:
                continue
            raise ConfigException(f"cfg {section=} was never used")
        found_options = used_cfgs[section]
        found_options = set(map(config2.optionxform, found_options))
        for option in config2.options(section):
            if option not in found_options:
                raise ConfigException(
                    f"cfg {section=} {option=} was never used")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
from datetime import datetime
import logging
import re
import sys
from typing import Any, List, Optional, Tuple
from inspect import getfullargspec
from .log_store import LogStore
from .overrides import overrides

_LEVELS = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL,
}


class ConfiguredFilter(object):
    """
    Allow a parent logger to filter a child logger.
    """
    __slots__ = [
        "_default_level", "_levels"]

    def __init__(self, conf):
        self._levels = ConfiguredFormatter.construct_logging_parents(conf)
        self._default_level = _LEVELS[conf.get("Logging", "default")]

    def filter(self, record):
        """
        Get the level for the deepest parent, and filter appropriately.
        """
        level = ConfiguredFormatter.level_of_deepest_parent(
            self._levels, record.name)

        if level is None:
            return record.levelno >= self._default_level

        return record.levelno >= level


class ConfiguredFormatter(logging.Formatter):
    """
    Defines the logging format for the SpiNNaker host software.
    """
    # Precompile this RE; it gets used quite a few times
    __last_component = re.compile(r'\.[^.]+$')

    def __init__(self, conf):
        if conf.get("Logging", "default") == "debug":
            fmt = "%(asctime)-15s %(levelname)s: %(pathname)s: %(message)s"
        else:
            fmt = "%(asctime)-15s %(levelname)s: %(message)s"
        super().__init__(fmt=fmt, datefmt="%Y-%m-%d %H:%M:%S")

    @staticmethod
    def construct_logging_parents(conf):
        """
        Create a dictionary of module names and logging levels.
        """
        # Construct the dictionary
        _levels = {}

        if not conf.has_section("Logging"):
            return _levels

        for label, level in _LEVELS.items():
            if conf.has_option("Logging", label):
                modules = [s.strip() for s in
                           conf.get('Logging', label).split(',')]
                if '' not in modules:
                    _levels.update(dict((m, level) for m in modules))
        return _levels

    @staticmethod
    def deepest_parent(parents, child):
        """
        Greediest match between child and parent.
        """
        # TODO: this can almost certainly be neater!
        # Repeatedly strip elements off the child until we match an item in
        # parents
        match = child

        while '.' in match and match not in parents:
            match = ConfiguredFormatter.__last_component.sub('', match)

        # If no match then return None, there is no deepest parent
        if match not in parents:
            match = None

        return match

    @staticmethod
    def level_of_deepest_parent(parents, child):
        """
        The logging level of the greediest match between child and parent.
        """
        # child = re.sub( r'^pacman103\.', '', child )
        parent = ConfiguredFormatter.deepest_parent(parents.keys(), child)

        if parent is None:
            return None

        return parents[parent]


class _BraceMessage(object):
    """
    A message that converts a Python format string to a string.
    """
    __slots__ = [
        "args", "fmt", "kwargs"]

    def __init__(self, fmt, args, kwargs):
        self.fmt = fmt
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        try:
            return str(self.fmt).format(*self.args, **self.kwargs)
        except KeyError:
            try:
                if (not self.args and not self.kwargs and
                        isinstance(self.fmt, str)):
                    # nothing to format with so assume brackets not formatting
                    return self.fmt
                return "KeyError: " + str(self.fmt)
            except KeyError:
                return "Double KeyError"
        except IndexError:
            try:
                if self.args or self.kwargs:
                    return "IndexError: " + str(self.fmt)
                else:
                    # nothing to format with so assume brackets not formatting
                    return str(self.fmt)
            except IndexError:
                return "Double IndexError"


class LogLevelTooHighException(Exception):
    """
    An Exception throw when the System tries to log at a level where an
    Exception is a better option.
    """


class FormatAdapter(logging.LoggerAdapter):
    """
    An adaptor for logging with messages that uses Python format strings.

    Example::

        log = FormatAdapter(logging.getLogger(__name__))
        log.info("this message has {} inside {}", 123, 'itself')
        # --> INFO: this message has 123 inside itself
    """
    __kill_level = logging.CRITICAL + 1
    __repeat_at_end = logging.WARNING
    __not_stored_messages: List[Tuple[datetime, int, str]] = []
    __log_store: Optional[LogStore] = None

    @classmethod
    def set_kill_level(cls, level: Optional[int] = None):
        """
        Allow system to change the level at which a log is changed to an
        Exception.

        .. note::
            This is a static method; it affects all log messages.

        :param int level:
            The level to set. The values in :py:mod:`logging` are recommended.
        """
        if level is None:
            cls.__kill_level = logging.CRITICAL + 1
        else:
            cls.__kill_level = level

    @classmethod
    def set_log_store(cls, log_store: Optional[LogStore]):
        """
        Sets a Object to write the log messages to
        :param LogStore log_store:
        """
        if log_store is not None and not isinstance(log_store, LogStore):
            raise TypeError("log_store must be a LogStore")
        cls.__log_store = log_store
        if cls.__log_store:
            for timestamp, level, message in cls._pop_not_stored_messages():
                cls.__log_store.store_log(level, message, timestamp)

    def __init__(self, logger: logging.Logger, extra=None):
        if extra is None:
            extra = {}
        super().__init__(logger, extra)
        self.do_log = logger._log  # pylint: disable=protected-access

    @overrides(logging.LoggerAdapter.log, extend_doc=False, adds_typing=True)
    def log(self, level: int, msg: object, *args, **kwargs):
        """
        Delegate a log call to the underlying logger, applying appropriate
        transformations to allow the log message to be written using
        Python format string, rather than via `%`-substitutions.
        """
        if level >= FormatAdapter.__kill_level:
            raise LogLevelTooHighException(_BraceMessage(msg, args, kwargs))
        if self.isEnabledFor(level):
            message = _BraceMessage(msg, args, kwargs)
            if FormatAdapter.__log_store:
                try:
                    FormatAdapter.__log_store.store_log(level, str(message))
                except Exception as ex:
                    # Avoid an endless loop of log store errors being logged
                    FormatAdapter.__not_stored_messages.append((
                        datetime.now(),
                        level,
                        f"Unable to store log messages in database due to"
                        f" {ex}"))
                    FormatAdapter.__not_stored_messages.append(
                        (datetime.now(), level, str(message)))
                    FormatAdapter.__log_store = None
                    raise
            else:
                FormatAdapter.__not_stored_messages.append(
                    (datetime.now(), level, str(message)))
            msg, log_kwargs = self.process(msg, kwargs)
            if "exc_info" in kwargs:
                log_kwargs["exc_info"] = kwargs["exc_info"]
            self.do_log(level, message, (), **log_kwargs)

    @overrides(logging.LoggerAdapter.process, extend_doc=False,
               adds_typing=True)
    def process(self, msg: object, kwargs: Any) -> Tuple[object, dict]:
        """
        Process the logging message and keyword arguments passed in to a
        logging call to insert contextual information. You can either
        manipulate the message itself, the keyword arguments or both.
        Return the message and *kwargs* modified (or not) to suit your needs.
        """
        return msg, {
            key: kwargs[key]
            for key in getfullargspec(self.do_log).args[1:]
            if key in kwargs}

    @classmethod
    def atexit_handler(cls) -> None:
        """
        Adds code to print out high level log messages python run ends
        """
        messages = []
        if cls.__log_store:
            try:
                messages = cls.__log_store.retreive_log_messages(
                    cls.__repeat_at_end)
            except Exception:  # pylint: disable=broad-except
                pass

        messages.extend(cls._pop_not_stored_messages(cls.__repeat_at_end))
        if messages:
            level = logging.getLevelName(cls.__repeat_at_end)
            print(f"\n!WARNING: {len(messages)} log messages were "
                  f"generated at level {level} or above.", file=sys.stderr)
            print("This may mean that the results are invalid.",
                  file=sys.stderr)
            if cls.__log_store:
                print(f"You are advised to check the details of these in "
                      f"the p_log_view of : "
                      f"{cls.__log_store.get_location()}", file=sys.stderr)
            if len(messages) < 10:
                print("These are:", file=sys.stderr)
            else:
                print("The first 10 are:", file=sys.stderr)
            for message in messages[0:10]:
                print(message, file=sys.stderr)

    @classmethod
    def _pop_not_stored_messages(cls, min_level=0):
        """
        Returns the log of messages to print on exit and
        *clears that log*.

        .. note::
            Should only be called externally from test code!
        """
        result = []
        try:
            for timestamp, level, message in cls.__not_stored_messages:
                if level >= min_level:
                    result.append((timestamp, level, message))
            return result
        except Exception:  # pylint: disable=broad-except
            return result
        finally:
            cls.__not_stored_messages = []


atexit.register(FormatAdapter.atexit_handler)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
from typing import List, Optional


NONES = ("none", )
TRUES = ('y', 'yes', 't', 'true', 'on', '1')
FALSES = ('n', 'no', 'f', 'false', 'off', '0')


class CamelCaseConfigParser(configparser.RawConfigParser):
    """
    Extends the Parser to allow for differences in case and underscores
    """
    __slots__ = ["_read_files"]

    def optionxform(self, optionstr: str) -> str:
        """
        Transforms the name of an option to lower case and strips
        underscores, so matching is more user-friendly.
        """
        lower = optionstr.lower()
        return lower.replace("_", "")

    def __init__(self, defaults=None):
        super().__init__(defaults)
        self._read_files = list()

    def read(self, filenames, encoding=None):
        """
        Read and parse a filename or a list of filenames.
        """
        new_files = super().read(filenames, encoding)
        self._read_files.extend(new_files)
        return new_files

    @property
    def read_files(self):
        """
        The configuration files that have been actually read.
        """
        return self._read_files

    def get_str(self, section: str, option: str) -> Optional[str]:
        """
        Get the string value of an option.

        :param str section: What section to get the option from.
        :param str option: What option to read.
        :return: The option value
        :rtype: str or None
        """
        value = self.get(section, option)
        if value.lower() in NONES:
            return None
        return value

    def get_str_list(
            self, section: str, option: str, token: str = ",") -> List[str]:
        """
        Get the string value of an option split into a list.

        :param str section: What section to get the option from.
        :param str option: What option to read.
        :param token: The token to split the string into a list
        :return: The list (possibly empty) of the option values
        :rtype: list(str)
        """
        value = self.get(section, option)
        if value.lower() in NONES:
            return []
        if len(value.strip()) == 0:
            return []
        as_list = value.split(token)
        return list(map(lambda x: x.strip(), as_list))

    def get_int(self, section: str, option: str) -> Optional[int]:
        """
        Get the integer value of an option.

        :param str section: What section to get the option from.
        :param str option: What option to read.
        :return: The option value
        :rtype: int
        """
        value = self.get(section, option)
        if str(value).lower() in NONES:
            return None
        return int(value)

    def get_float(self, section: str, option: str) -> Optional[float]:
        """
        Get the float value of an option.

        :param str section: What section to get the option from.
        :param str option: What option to read.
        :return: The option value.
        :rtype: float
        """
        value = self.get(section, option)
        if str(value).lower() in NONES:
            return None
        return float(value)

    def get_bool(self, section: str, option: str) -> Optional[bool]:
        """
        Get the Boolean value of an option.

        :param str section: What section to get the option from.
        :param str option: What option to read.
        :return: The option value.
        :rtype: bool
        """
        value = self.get(section, option)
        lower = str(value).lower()
        if lower in TRUES:
            return True
        if lower in FALSES:
            return False
        if lower in NONES:
            return None
        raise ValueError(f"invalid truth value {value}")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from configparser import NoOptionError
import logging
import os
from typing import Callable, Dict, List, Sequence, Tuple, Union

import appdirs
from typing_extensions import TypeAlias

from spinn_utilities import log
from spinn_utilities.configs import (
    CamelCaseConfigParser, ConfigTemplateException,
    NoConfigFoundException, UnexpectedConfigException)
logger = log.FormatAdapter(logging.getLogger(__name__))
_SectionParser: TypeAlias = Callable[[CamelCaseConfigParser], None]


def install_cfg_and_error(
        filename: str, defaults: List[str],
        config_locations: List[str]) -> NoConfigFoundException:
    """
    Installs a local configuration file based on the templates and raises
    an exception.

    This method is called when no user configuration file is found.

    It will create a file in the users home directory based on the defaults.
    Then it prints a helpful message and throws an error with the same message.

    :param str filename:
        Name under which to save the new configuration file
    :param list(str) defaults:
        List of full paths to the default configuration files.
        Each of which *must* have an associated template file with exactly the
        same path plus `.template`.
    :param list(str) config_locations:
        List of paths where the user configuration files
        were looked for. Only used for the message
    :raise spinn_utilities.configs.NoConfigFoundException:
        Always raised
    """
    home_cfg = os.path.join(os.path.expanduser("~"), f".{filename}")

    found = False
    with open(home_cfg, "w", encoding="utf-8") as dst:
        for source in defaults:
            template = source + ".template"
            if os.path.isfile(template):
                if found:
                    raise ConfigTemplateException(
                        f"Second template found at {template}")
                with open(source + ".template", "r", encoding="utf-8") as src:
                    dst.write(src.read())
                    dst.write("\n")
                    found = True
        if not found:
            if defaults:
                raise ConfigTemplateException(
                    f"No template file found for {defaults}")
            else:
                raise ConfigTemplateException(
                    f"No default cfg files found. "
                    f"New {home_cfg} will be empty")

        dst.write("\n# Additional config options can be found in:\n")
        for source in defaults:
            dst.write(f"# {source}\n")
        dst.write("\n# Copy any additional settings you want to change"
                  " here including section headings\n")

    msg = (f'Unable to find config file in any of the following locations: \n'
           f'{config_locations}\n'
           f'**********************************************************\n'
           f'{home_cfg} has been created. \n'
           f'Please edit this file and change "None" after "machineName" '
           f'to the hostname or IP address of your SpiNNaker board, '
           f'and change "None" after "version" to the version of '
           f'SpiNNaker hardware you are running on:\n'
           f'[Machine]\n'
           f'machineName = None\n'
           f'version = None\n'
           f'***********************************************************\n')
    print(msg)
    return NoConfigFoundException(msg)


def logging_parser(config: CamelCaseConfigParser):
    """
    Create the root logger with the given level.

    Create filters based on logging levels

    .. note::
        You do not normally need to call this function; it is used
        automatically to parse Logging configuration sections.
    """
    try:
        if config.getboolean("Logging", "instantiate"):
            level = config.get("Logging", "default").upper()
            logging.basicConfig(level=level)
        for handler in logging.root.handlers:
            handler.addFilter(
                log.ConfiguredFilter(config))  # type: ignore[arg-type]
            handler.setFormatter(log.ConfiguredFormatter(config))
    except NoOptionError:
        pass


def _check_config(
        cfg_file: str, default_configs: CamelCaseConfigParser, strict: bool):
    """
    Checks the configuration read up to this point to see if it is outdated.

    Once one difference is found a full reports is generated and an error
    raised.

    Any section specifically listed as Dead will cause a error

    Any section in the default_cfg should not have extra values.
    It will never have less as the default_cfg are in the configuration.

    Errors on any values listed as PreviousValues.
    These are specific values in specific options no longer supported.
    For example old algorithm names.

    :param str cfg_file: Path of last file read in
    :param CamelCaseConfigParser default_configs:
        configuration with just the default files in
    :param bool strict: Flag to say an exception should be raised
    """
    if not default_configs.sections():  # empty
        logger.warning("Can not validate cfg files as no default.")
        return
    configs = CamelCaseConfigParser()
    configs.read(cfg_file)
    msg = ""
    for section in configs.sections():
        if default_configs.has_section(section):
            for option in configs.options(section):
                if not default_configs.has_option(section, option):
                    msg += f"Unexpected Option: [{section}]{option}\n"
        else:
            msg += f"Unexpected Section: [{section}]\n"
    if msg:
        msg += f"found in {cfg_file}"
        if strict:
            raise UnexpectedConfigException(msg)
        else:
            logger.warning(msg)


def _read_a_config(
        configuration: CamelCaseConfigParser, cfg_file: str,
        default_configs: CamelCaseConfigParser, strict: bool):
    """
    Reads in a configuration file and then directly its `machine_spec_file`.

    :param CamelCaseConfigParser configuration:
        configuration to be updated by the reading of a file
    :param str cfg_file: path to file which should be read in
    :param CamelCaseConfigParser default_configs:
        configuration with just the default files in
    :param bool strict: Flag to say checker should raise an exception
    """
    _check_config(cfg_file, default_configs, strict)
    configuration.read(cfg_file)
    if configuration.has_option("Machine", "machine_spec_file"):
        machine_spec_file = configuration.get("Machine", "machine_spec_file")
        _check_config(machine_spec_file, default_configs, strict)
        configuration.read(machine_spec_file)
        configuration.remove_option("Machine", "machine_spec_file")


def _config_locations(filename: str) -> List[str]:
    """
    Defines the list of places we can get configuration files from.

    :param str filename:
        The local name of the configuration file, e.g., 'spynnaker.cfg'
    :return: list of fully-qualified filenames
    :rtype: list(str)
    """
    dotname = "." + filename

    # locations to read as well as default later overrides earlier
    system_config_cfg_file = os.path.join(appdirs.site_config_dir(), dotname)
    user_config_cfg_file = os.path.join(appdirs.user_config_dir(), dotname)
    user_home_cfg_file = os.path.join(os.path.expanduser("~"), dotname)

    # locations to read as well as default later overrides earlier
    return [system_config_cfg_file, user_config_cfg_file,
            user_home_cfg_file]


def load_config(
        filename: str, defaults: List[str], config_parsers: Union[
            Sequence[Tuple[str, _SectionParser]],
            Dict[str, _SectionParser]] = ()) -> CamelCaseConfigParser:
    """
    Load the configuration.

    :param str filename:
        The base name of the configuration file(s).
        Should not include any path components.
    :param list(str) defaults:
        The list of files to get default configurations from.
    :param config_parsers:
        The parsers to parse the sections of the configuration file with, as
        a list of (section name, parser) or a dictionary from section name to
        parser; a configuration section will only
        be parsed if the section_name is found in the configuration files
        already loaded. The standard logging parser is appended to (a copy
        of) this.
    :type config_parsers: list(tuple(str, ~configparser.RawConfigParser))
    :return: the fully-loaded and checked configuration
    :rtype: ~configparser.RawConfigParser
    """
    configs = CamelCaseConfigParser()

    # locations to read as well as default later overrides earlier
    config_locations = _config_locations(filename)
    if not any(os.path.isfile(f) for f in config_locations):
        if defaults:
            raise install_cfg_and_error(
                filename, defaults, config_locations)
        else:
            logger.error("No default cfg files provided")

    configs.read(defaults)

    default_configs = CamelCaseConfigParser()
    default_configs.read(defaults)

    for cfg_file in config_locations:
        _read_a_config(configs, cfg_file, default_configs, False)
    cfg_file = os.path.join(os.curdir, filename)
    _read_a_config(configs, cfg_file, default_configs, True)

    parsers = dict(config_parsers)
    parsers["Logging"] = logging_parser

    for section in parsers:
        if configs.has_section(section):
            parsers[section](configs)

    # Log which configs files we read
    print(configs.read_files)
    logger.info("Read configs files: {}", ", ".join(configs.read_files))

    return configs



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from spinn_utilities.config_holder import (
    add_default_cfg, clear_cfg_files)
from spinn_utilities.data.utils_data_writer import UtilsDataWriter
from spinn_utilities.overrides import overrides

BASE_CONFIG_FILE = "spinn_utilities.cfg"


def unittest_setup() -> None:
    """
    Resets the configurations so only the local default configuration is
    included.

    .. note::
        This file should only be called from `SpiNNUtils/unittests`

    :param unittest_mode: Flag to indicate in unit tests
    """
    clear_cfg_files(True)
    add_spinn_utilities_cfg()
    UtilsDataWriter.mock()


def add_spinn_utilities_cfg() -> None:
    """
    Loads the default config values for spinn_utilities
    """
    overrides.check_types()
    add_default_cfg(os.path.join(os.path.dirname(__file__), BASE_CONFIG_FILE))



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import os.path
import logging
from typing_extensions import Self
from spinn_utilities.exceptions import (
    IllegalWriterException, InvalidDirectory, SimulatorNotRunException,
    UnexpectedStateChange)
from spinn_utilities.executable_finder import ExecutableFinder
from spinn_utilities.log import FormatAdapter
from .data_status import DataStatus
from .reset_status import ResetStatus
from .run_status import RunStatus
from .utils_data_view import UtilsDataView, _UtilsDataModel

logger = FormatAdapter(logging.getLogger(__file__))
# pylint: disable=protected-access


class UtilsDataWriter(UtilsDataView):
    """
    Writer class for the information in UtilsDataView

    This and subclass Writers are specifically designed to be used by the
    simulators (ASB) and unit tests only!
    Any other usage is not supported.

    The Writers are specifically designed to support only one instant
    (typically held by ASB as `self._data_writer`).

    Creating a new instant of the Writer will clear out any data added by the
    previous instance.

    Unit tests can create a writer by doing `...Writer.mock()` or
    `Writer.setup()`.
    The `mock` method adds some default data such as directories and
    a Virtual 8 * 8 Machine, as well as allowing some back-door methods.
    `...Writer.mock()` is the recommended one for unit tests;
    `setup()` is more like what ASB does and allows for state changes such as
    `writer.start_running`.

    ASB `__init__()` (or it subclasses) will create a new writer
    so a call to `sim.setup` will clear all previously held data.

    .. warning::
        As the Writers are not designed for general usage the methods can
        change without notice.
    """
    __data = _UtilsDataModel()
    __slots__ = ()

    def __init__(self, state: DataStatus):
        """
        :param ~spinn_utilities.data.DataStatus state:
            State writer should be in
        """
        if state == DataStatus.MOCKED:
            self._mock()
        elif state == DataStatus.SETUP:
            self._setup()
        else:
            raise IllegalWriterException(
                "Writers can only be created by mock or setup")

    @classmethod
    def mock(cls) -> Self:
        """
        Creates a writer in mock mode.

        All previous data will be cleared.

        This should set the most likely defaults values.
        But be aware that what is considered the most likely default could
        change over time.

        Unit tests that depend on any valid value being set should be able to
        depend on Mock.

        A unit test that depends on a specific value should call `mock()` and
        then set that value.

        :return: A Data Writer
        :rtype: UtilsDataWriter
        """
        return cls(DataStatus.MOCKED)

    @classmethod
    def setup(cls) -> Self:
        """
        Creates a writer in normal mode.

        All previous data will be cleared

        :return: A Data Writer
        :rtype: UtilsDataWriter
        """
        return cls(DataStatus.SETUP)

    def _mock(self) -> None:
        """
        This method should only be called by `mock` (via `__init__`).
        """
        self.__data._clear()
        self.__data._data_status = DataStatus.MOCKED
        self.__data._reset_status = ResetStatus.NOT_SETUP
        self.__data._run_status = RunStatus.NOT_SETUP

    def _setup(self) -> None:
        """
        This method should only be called by `setup` (via `__init__`).
        """
        self.__data._clear()
        self.__data._data_status = DataStatus.SETUP
        self.__data._reset_status = ResetStatus.SETUP
        self.__data._run_status = RunStatus.NOT_RUNNING

    def start_run(self) -> None:
        """
        Puts all data into the state expected after `do_run_loop` started.
        """
        if self.__data._run_status != RunStatus.NOT_RUNNING:
            self.check_valid_simulator()
            raise UnexpectedStateChange(
                f"Unexpected start run when in run state "
                f"{self.__data._run_status}")
        self.__data._run_status = RunStatus.IN_RUN

    def finish_run(self) -> None:
        """
        Puts all data into the state expected after `sim.run` ends.
        """
        if self.__data._run_status not in [
                RunStatus.IN_RUN, RunStatus.STOP_REQUESTED]:
            self.check_valid_simulator()
            raise UnexpectedStateChange(
                f"Unexpected finish run when in run state "
                f"{self.__data._run_status}")
        self.__data._run_status = RunStatus.NOT_RUNNING
        self.__data._reset_status = ResetStatus.HAS_RUN
        self.__data._requires_data_generation = False
        self.__data._requires_mapping = False

    def _hard_reset(self) -> None:
        """
        This method should only be called by `hard_setup`.
        """
        self.__data._hard_reset()
        self.__data._reset_status = ResetStatus.HARD_RESET

    def hard_reset(self) -> None:
        """
        Puts all data back into the state expected at graph changed and
        `sim.reset`.

        This resets any data set after `sim.setup` has finished.
        """
        if self.__data._run_status not in [
                RunStatus.IN_RUN, RunStatus.STOP_REQUESTED]:
            self.check_user_can_act()
        if self.__data._reset_status in [
                ResetStatus.HAS_RUN,
                ResetStatus.SOFT_RESET]:
            # call the protected method at the highest possible level
            self._hard_reset()
            return
        self.check_valid_simulator()
        if self.__data._reset_status == ResetStatus.SETUP:
            raise SimulatorNotRunException(
                "Calling reset before calling run is not supported")
        raise UnexpectedStateChange(
            f"Unexpected call to reset while reset status is "
            f"{self.__data._reset_status}")

    def _soft_reset(self) -> None:
        """
        This method should only be called from `soft_reset`.
        """
        self.__data._soft_reset()
        self.__data._reset_status = ResetStatus.SOFT_RESET

    def soft_reset(self) -> None:
        """
        Puts all data back into the state expected at `sim.reset` but where the
        graph has not changed.
        """
        self.check_user_can_act()
        if self.__data._reset_status == ResetStatus.HAS_RUN:
            # call the protected method at the highest possible level
            self._soft_reset()
            return
        self.check_valid_simulator()
        if self.__data._reset_status == ResetStatus.SETUP:
            raise SimulatorNotRunException(
                "Calling reset before calling run is not supported")
        raise UnexpectedStateChange(
            f"Unexpected call to reset while reset status is "
            f"{self.__data._reset_status}")

    def request_stop(self) -> None:
        """
        Used to indicate a user has requested a stop.

        This is expected to be called during `run` from a different thread.
        """
        if self.__data._run_status != RunStatus.IN_RUN:
            self.check_valid_simulator()
            raise UnexpectedStateChange(
                f"Unexpected request stop when in run state "
                f"{self.__data._run_status}")
        self.__data._run_status = RunStatus.STOP_REQUESTED

    def stopping(self) -> None:
        """
        Puts all data into the state expected during stop.

        :raises SimulatorNotSetupException: If called before `sim.setup`
        :raises SimulatorShutdownException: If called after `sim.end`
        """
        if self.__data._run_status not in [
                RunStatus.NOT_RUNNING, RunStatus.IN_RUN,
                RunStatus.STOP_REQUESTED]:
            self.check_valid_simulator()
            raise UnexpectedStateChange(
                "Unexpected call to stopping while in run_state"
                f" {self.__data._run_status}")
        self.__data._run_status = RunStatus.STOPPING

    def shut_down(self) -> None:
        """
        Puts all data into the state expected after `sim.end`.

        Most methods that change data, or state will raise an exception after
        this call.

        Most data however will still be available.
        """
        self.__data._data_status = DataStatus.SHUTDOWN
        self.__data._run_status = RunStatus.SHUTDOWN

    def get_report_dir_path(self) -> str:
        """
        Returns path to existing reports directory.

        This is the high level directory which in which `timestamp` directories
        and `run` directories are placed.

        As it is only accessed to create `timestamp` directories and
        remove old reports, this is not a view method.

        :rtype: str
        :raises SpiNNUtilsException:
            If the `simulation_time_step` is currently unavailable
        """
        if self.__data._report_dir_path:
            return self.__data._report_dir_path
        raise self._exception("report_dir_path")

    def set_run_dir_path(self, run_dir_path: str):
        """
        Checks and sets the `run_dir_path`.

        :param str run_dir_path:
        :raises InvalidDirectory: if the `run_dir_path` is not a directory
        """
        if os.path.isdir(run_dir_path):
            self.__data._run_dir_path = run_dir_path
        else:
            self.__data._run_dir_path = None
            raise InvalidDirectory("run_dir_path", run_dir_path)

    def set_report_dir_path(self, reports_dir_path: str):
        """
        Checks and sets the `reports_dir_path`.

        :param str reports_dir_path:
        :raises InvalidDirectory: if the `reports_dir_path` is not a directory
        """
        if os.path.isdir(reports_dir_path):
            self.__data._report_dir_path = reports_dir_path
        else:
            self.__data._report_dir_path = None
            raise InvalidDirectory("run_dir_path", reports_dir_path)

    def _set_executable_finder(self, executable_finder: ExecutableFinder):
        """
        Only usable by unit tests!

        :param ExecutableFinder executable_finder:
        """
        if not self._is_mocked():
            raise NotImplementedError("Only valid in Mocked state!")
        self.__data._executable_finder = executable_finder
